@@ -55,7 +55,7 @@ enum class JpegEncoder {
 // Popular jpeg scan scripts
 // The fields of the individual scans are:
 // comps_in_scan, component_index[], Ss, Se, Ah, Al
-static constexpr jpeg_scan_info kScanScript1[] = {
+constexpr jpeg_scan_info kScanScript1[] = {
     {1, {0}, 0, 0, 0, 0},   //
     {1, {1}, 0, 0, 0, 0},   //
     {1, {2}, 0, 0, 0, 0},   //
@@ -64,9 +64,9 @@ static constexpr jpeg_scan_info kScanScript1[] = {
     {1, {1}, 1, 63, 0, 0},  //
     {1, {2}, 1, 63, 0, 0},  //
 };
-static constexpr size_t kNumScans1 = ARRAY_SIZE(kScanScript1);
+constexpr size_t kNumScans1 = ARRAY_SIZE(kScanScript1);
 
-static constexpr jpeg_scan_info kScanScript2[] = {
+constexpr jpeg_scan_info kScanScript2[] = {
     {1, {0}, 0, 0, 0, 0},   //
     {1, {1}, 0, 0, 0, 0},   //
     {1, {2}, 0, 0, 0, 0},   //
@@ -76,9 +76,9 @@ static constexpr jpeg_scan_info kScanScript2[] = {
     {1, {1}, 1, 63, 0, 0},  //
     {1, {2}, 1, 63, 0, 0},  //
 };
-static constexpr size_t kNumScans2 = ARRAY_SIZE(kScanScript2);
+constexpr size_t kNumScans2 = ARRAY_SIZE(kScanScript2);
 
-static constexpr jpeg_scan_info kScanScript3[] = {
+constexpr jpeg_scan_info kScanScript3[] = {
     {1, {0}, 0, 0, 0, 0},   //
     {1, {1}, 0, 0, 0, 0},   //
     {1, {2}, 0, 0, 0, 0},   //
@@ -88,9 +88,9 @@ static constexpr jpeg_scan_info kScanScript3[] = {
     {1, {1}, 1, 63, 0, 0},  //
     {1, {2}, 1, 63, 0, 0},  //
 };
-static constexpr size_t kNumScans3 = ARRAY_SIZE(kScanScript3);
+constexpr size_t kNumScans3 = ARRAY_SIZE(kScanScript3);
 
-static constexpr jpeg_scan_info kScanScript4[] = {
+constexpr jpeg_scan_info kScanScript4[] = {
     {3, {0, 1, 2}, 0, 0, 0, 1},  //
     {1, {0}, 1, 5, 0, 2},        //
     {1, {2}, 1, 63, 0, 1},       //
@@ -102,9 +102,9 @@ static constexpr jpeg_scan_info kScanScript4[] = {
     {1, {1}, 1, 63, 1, 0},       //
     {1, {0}, 1, 63, 1, 0},       //
 };
-static constexpr size_t kNumScans4 = ARRAY_SIZE(kScanScript4);
+constexpr size_t kNumScans4 = ARRAY_SIZE(kScanScript4);
 
-static constexpr jpeg_scan_info kScanScript5[] = {
+constexpr jpeg_scan_info kScanScript5[] = {
     {3, {0, 1, 2}, 0, 0, 0, 1},  //
     {1, {0}, 1, 5, 0, 2},        //
     {1, {1}, 1, 5, 0, 2},        //
@@ -120,10 +120,10 @@ static constexpr jpeg_scan_info kScanScript5[] = {
     {1, {1}, 1, 63, 1, 0},       //
     {1, {2}, 1, 63, 1, 0},       //
 };
-static constexpr size_t kNumScans5 = ARRAY_SIZE(kScanScript5);
+constexpr size_t kNumScans5 = ARRAY_SIZE(kScanScript5);
 
 // default progressive mode of jpegli
-static constexpr jpeg_scan_info kScanScript6[] = {
+constexpr jpeg_scan_info kScanScript6[] = {
     {3, {0, 1, 2}, 0, 0, 0, 0},  //
     {1, {0}, 1, 2, 0, 0},        //
     {1, {1}, 1, 2, 0, 0},        //
@@ -138,7 +138,7 @@ static constexpr jpeg_scan_info kScanScript6[] = {
     {1, {1}, 3, 63, 1, 0},       //
     {1, {2}, 3, 63, 1, 0},       //
 };
-static constexpr size_t kNumScans6 = ARRAY_SIZE(kScanScript6);
+constexpr size_t kNumScans6 = ARRAY_SIZE(kScanScript6);
 
 // Adapt RGB scan info to grayscale jpegs.
 void FilterScanComponents(const jpeg_compress_struct* cinfo,
@@ -186,14 +186,6 @@ Status SetJpegProgression(int progressive_id,
   return true;
 }
 
-bool IsSRGBEncoding(const JxlColorEncoding& c) {
-  return ((c.color_space == JXL_COLOR_SPACE_RGB ||
-           c.color_space == JXL_COLOR_SPACE_GRAY) &&
-          c.primaries == JXL_PRIMARIES_SRGB &&
-          c.white_point == JXL_WHITE_POINT_D65 &&
-          c.transfer_function == JXL_TRANSFER_FUNCTION_SRGB);
-}
-
 void WriteICCProfile(jpeg_compress_struct* const cinfo,
                      const std::vector<uint8_t>& icc) {
   constexpr size_t kMaxIccBytesInMarker =
@@ -292,7 +284,7 @@ Status EncodeWithLibJpeg(const PackedImage& image, const JxlBasicInfo& info,
   cinfo.input_components = info.num_color_channels;
   cinfo.in_color_space = info.num_color_channels == 1 ? JCS_GRAYSCALE : JCS_RGB;
   jpeg_set_defaults(&cinfo);
-  cinfo.optimize_coding = params.optimize_coding;
+  cinfo.optimize_coding = static_cast<boolean>(params.optimize_coding);
   if (cinfo.input_components == 3) {
     JXL_RETURN_IF_ERROR(
         SetChromaSubsampling(params.chroma_subsampling, &cinfo));
@@ -318,7 +310,7 @@ Status EncodeWithLibJpeg(const PackedImage& image, const JxlBasicInfo& info,
 
   std::vector<uint8_t> row_bytes(image.stride);
   const uint8_t* pixels = reinterpret_cast<const uint8_t*>(image.pixels());
-  if (cinfo.num_components == (int)image.format.num_channels &&
+  if (cinfo.num_components == static_cast<int>(image.format.num_channels) &&
       image.format.data_type == JXL_TYPE_UINT8) {
     for (size_t y = 0; y < info.ysize; ++y) {
       memcpy(&row_bytes[0], pixels + y * image.stride, image.stride);
@@ -598,18 +590,14 @@ class JPEGEncoder : public Encoder {
       }
     }
     params.is_xyb = (ppf.color_encoding.color_space == JXL_COLOR_SPACE_XYB);
-    std::vector<uint8_t> icc;
-    if (!IsSRGBEncoding(ppf.color_encoding)) {
-      icc = ppf.icc;
-    }
     encoded_image->bitstreams.clear();
     encoded_image->bitstreams.reserve(ppf.frames.size());
     for (const auto& frame : ppf.frames) {
       JXL_RETURN_IF_ERROR(VerifyPackedImage(frame.color, ppf.info));
       encoded_image->bitstreams.emplace_back();
       JXL_RETURN_IF_ERROR(EncodeImageJPG(
-          frame.color, ppf.info, icc, ppf.metadata.exif, jpeg_encoder, params,
-          pool, &encoded_image->bitstreams.back()));
+          frame.color, ppf.info, ppf.icc, ppf.metadata.exif, jpeg_encoder,
+          params, pool, &encoded_image->bitstreams.back()));
     }
     return true;
   }
