@@ -9,6 +9,7 @@
 #include <stdlib.h>
 
 #include <algorithm>
+#include <atomic>
 #include <cmath>
 #include <string>
 #include <vector>
@@ -26,8 +27,6 @@
 #include "lib/jxl/base/status.h"
 #include "lib/jxl/butteraugli/butteraugli.h"
 #include "lib/jxl/cms/opsin_params.h"
-#include "lib/jxl/coeff_order_fwd.h"
-#include "lib/jxl/color_encoding_internal.h"
 #include "lib/jxl/convolve.h"
 #include "lib/jxl/dec_cache.h"
 #include "lib/jxl/dec_group.h"
@@ -41,7 +40,6 @@
 #include "lib/jxl/enc_transforms-inl.h"
 #include "lib/jxl/epf.h"
 #include "lib/jxl/frame_dimensions.h"
-#include "lib/jxl/gauss_blur.h"
 #include "lib/jxl/image.h"
 #include "lib/jxl/image_bundle.h"
 #include "lib/jxl/image_ops.h"
@@ -103,12 +101,12 @@ V ComputeMask(const D d, const V out_val) {
 }
 
 // mul and mul2 represent a scaling difference between jxl and butteraugli.
-static const float kSGmul = 226.77216153508914f;
-static const float kSGmul2 = 1.0f / 73.377132366608819f;
-static const float kLog2 = 0.693147181f;
+const float kSGmul = 226.77216153508914f;
+const float kSGmul2 = 1.0f / 73.377132366608819f;
+const float kLog2 = 0.693147181f;
 // Includes correction factor for std::log -> log2.
-static const float kSGRetMul = kSGmul2 * 18.6580932135f * kLog2;
-static const float kSGVOffset = 7.7825991679894591f;
+const float kSGRetMul = kSGmul2 * 18.6580932135f * kLog2;
+const float kSGVOffset = 7.7825991679894591f;
 
 template <bool invert, typename D, typename V>
 V RatioOfDerivativesOfCubicRootToSimpleGamma(const D d, V v) {
@@ -132,7 +130,7 @@ V RatioOfDerivativesOfCubicRootToSimpleGamma(const D d, V v) {
 }
 
 template <bool invert = false>
-static float RatioOfDerivativesOfCubicRootToSimpleGamma(float v) {
+float RatioOfDerivativesOfCubicRootToSimpleGamma(float v) {
   using DScalar = HWY_CAPPED(float, 1);
   auto vscalar = Load(DScalar(), &v);
   return GetLane(
@@ -164,7 +162,8 @@ V SimpleGamma(const D d, V v) {
 
 template <class D, class V>
 V GammaModulation(const D d, const size_t x, const size_t y,
-                  const ImageF& xyb_x, const ImageF& xyb_y, const V out_val) {
+                  const ImageF& xyb_x, const ImageF& xyb_y, const Rect& rect,
+                  const V out_val) {
   const float kBias = 0.16f;
   JXL_DASSERT(kBias > jxl::cms::kOpsinAbsorbanceBias[0]);
   JXL_DASSERT(kBias > jxl::cms::kOpsinAbsorbanceBias[1]);
@@ -173,8 +172,8 @@ V GammaModulation(const D d, const size_t x, const size_t y,
   auto bias = Set(d, kBias);
   auto half = Set(d, 0.5f);
   for (size_t dy = 0; dy < 8; ++dy) {
-    const float* const JXL_RESTRICT row_in_x = xyb_x.Row(y + dy);
-    const float* const JXL_RESTRICT row_in_y = xyb_y.Row(y + dy);
+    const float* const JXL_RESTRICT row_in_x = rect.ConstRow(xyb_x, y + dy);
+    const float* const JXL_RESTRICT row_in_y = rect.ConstRow(xyb_y, y + dy);
     for (size_t dx = 0; dx < 8; dx += Lanes(d)) {
       const auto iny = Add(Load(d, row_in_y + x + dx), bias);
       const auto inx = Load(d, row_in_x + x + dx);
@@ -201,7 +200,7 @@ V GammaModulation(const D d, const size_t x, const size_t y,
 // Change precision in 8x8 blocks that have high frequency content.
 template <class D, class V>
 V HfModulation(const D d, const size_t x, const size_t y, const ImageF& xyb,
-               const V out_val) {
+               const Rect& rect, const V out_val) {
   // Zero out the invalid differences for the rightmost value per row.
   const Rebind<uint32_t, D> du;
   HWY_ALIGN constexpr uint32_t kMaskRight[kBlockDim] = {~0u, ~0u, ~0u, ~0u,
@@ -212,9 +211,9 @@ V HfModulation(const D d, const size_t x, const size_t y, const ImageF& xyb,
   static const float valmin = 0.020602694503245016f;
   auto valminv = Set(d, valmin);
   for (size_t dy = 0; dy < 8; ++dy) {
-    const float* JXL_RESTRICT row_in = xyb.Row(y + dy) + x;
+    const float* JXL_RESTRICT row_in = rect.ConstRow(xyb, y + dy) + x;
     const float* JXL_RESTRICT row_in_next =
-        dy == 7 ? row_in : xyb.Row(y + dy + 1) + x;
+        dy == 7 ? row_in : rect.ConstRow(xyb, y + dy + 1) + x;
 
     // In SCALAR, there is no guarantee of having extra row padding.
     // Hence, we need to ensure we don't access pixels outside the row itself.
@@ -252,11 +251,8 @@ V HfModulation(const D d, const size_t x, const size_t y, const ImageF& xyb,
 
 void PerBlockModulations(const float butteraugli_target, const ImageF& xyb_x,
                          const ImageF& xyb_y, const ImageF& xyb_b,
-                         const float scale, const Rect& rect, ImageF* out) {
-  JXL_ASSERT(SameSize(xyb_x, xyb_y));
-  JXL_ASSERT(DivCeil(xyb_x.xsize(), kBlockDim) == out->xsize());
-  JXL_ASSERT(DivCeil(xyb_x.ysize(), kBlockDim) == out->ysize());
-
+                         const Rect& rect_in, const float scale,
+                         const Rect& rect_out, ImageF* out) {
   float base_level = 0.48f * scale;
   float kDampenRampStart = 2.0f;
   float kDampenRampEnd = 14.0f;
@@ -270,16 +266,16 @@ void PerBlockModulations(const float butteraugli_target, const ImageF& xyb_x,
   }
   const float mul = scale * dampen;
   const float add = (1.0f - dampen) * base_level;
-  for (size_t iy = rect.y0(); iy < rect.y0() + rect.ysize(); iy++) {
+  for (size_t iy = rect_out.y0(); iy < rect_out.y1(); iy++) {
     const size_t y = iy * 8;
     float* const JXL_RESTRICT row_out = out->Row(iy);
     const HWY_CAPPED(float, kBlockDim) df;
-    for (size_t ix = rect.x0(); ix < rect.x0() + rect.xsize(); ix++) {
+    for (size_t ix = rect_out.x0(); ix < rect_out.x1(); ix++) {
       size_t x = ix * 8;
       auto out_val = Set(df, row_out[ix]);
       out_val = ComputeMask(df, out_val);
-      out_val = HfModulation(df, x, y, xyb_y, out_val);
-      out_val = GammaModulation(df, x, y, xyb_x, xyb_y, out_val);
+      out_val = HfModulation(df, x, y, xyb_y, rect_in, out_val);
+      out_val = GammaModulation(df, x, y, xyb_x, xyb_y, rect_in, out_val);
       // We want multiplicative quantization field, so everything
       // until this point has been modulating the exponent.
       row_out[ix] = FastPow2f(GetLane(out_val) * 1.442695041f) * mul + add;
@@ -399,24 +395,23 @@ void FuzzyErosion(const float butteraugli_target, const Rect& from_rect,
 }
 
 struct AdaptiveQuantizationImpl {
-  void Init(const Image3F& xyb) {
-    JXL_DASSERT(xyb.xsize() % kBlockDim == 0);
-    JXL_DASSERT(xyb.ysize() % kBlockDim == 0);
-    const size_t xsize = xyb.xsize();
-    const size_t ysize = xyb.ysize();
-    aq_map = ImageF(xsize / kBlockDim, ysize / kBlockDim);
-  }
-  void PrepareBuffers(size_t num_threads) {
-    diff_buffer = ImageF(kEncTileDim + 8, num_threads);
+  Status PrepareBuffers(size_t num_threads) {
+    JXL_ASSIGN_OR_RETURN(diff_buffer,
+                         ImageF::Create(kEncTileDim + 8, num_threads));
     for (size_t i = pre_erosion.size(); i < num_threads; i++) {
-      pre_erosion.emplace_back(kEncTileDimInBlocks * 2 + 2,
-                               kEncTileDimInBlocks * 2 + 2);
+      JXL_ASSIGN_OR_RETURN(ImageF tmp,
+                           ImageF::Create(kEncTileDimInBlocks * 2 + 2,
+                                          kEncTileDimInBlocks * 2 + 2));
+      pre_erosion.emplace_back(std::move(tmp));
     }
+    return true;
   }
 
   void ComputeTile(float butteraugli_target, float scale, const Image3F& xyb,
-                   const Rect& rect, const int thread, ImageF* mask,
-                   ImageF* mask1x1) {
+                   const Rect& rect_in, const Rect& rect_out, const int thread,
+                   ImageF* mask, ImageF* mask1x1) {
+    JXL_ASSERT(rect_in.x0() % 8 == 0);
+    JXL_ASSERT(rect_in.y0() % 8 == 0);
     const size_t xsize = xyb.xsize();
     const size_t ysize = xyb.ysize();
 
@@ -429,21 +424,30 @@ struct AdaptiveQuantizationImpl {
 
     const HWY_FULL(float) df;
 
-    size_t y_start = rect.y0() * 8;
-    size_t y_end = y_start + rect.ysize() * 8;
+    size_t y_start_1x1 = rect_in.y0() + rect_out.y0() * 8;
+    size_t y_end_1x1 = y_start_1x1 + rect_out.ysize() * 8;
 
-    size_t x_start = rect.x0() * 8;
-    size_t x_end = x_start + rect.xsize() * 8;
+    size_t x_start_1x1 = rect_in.x0() + rect_out.x0() * 8;
+    size_t x_end_1x1 = x_start_1x1 + rect_out.xsize() * 8;
+
+    if (rect_in.x0() != 0 && rect_out.x0() == 0) x_start_1x1 -= 2;
+    if (rect_in.x1() < xsize && rect_out.x1() * 8 == rect_in.xsize()) {
+      x_end_1x1 += 2;
+    }
+    if (rect_in.y0() != 0 && rect_out.y0() == 0) y_start_1x1 -= 2;
+    if (rect_in.y1() < ysize && rect_out.y1() * 8 == rect_in.ysize()) {
+      y_end_1x1 += 2;
+    }
 
     // Computes image (padded to multiple of 8x8) of local pixel differences.
     // Subsample both directions by 4.
     // 1x1 Laplacian of intensity.
-    for (size_t y = y_start; y < y_end; ++y) {
+    for (size_t y = y_start_1x1; y < y_end_1x1; ++y) {
       const size_t y2 = y + 1 < ysize ? y + 1 : y;
       const size_t y1 = y > 0 ? y - 1 : y;
-      const float* row_in = xyb.PlaneRow(1, y);
-      const float* row_in1 = xyb.PlaneRow(1, y1);
-      const float* row_in2 = xyb.PlaneRow(1, y2);
+      const float* row_in = xyb.ConstPlaneRow(1, y);
+      const float* row_in1 = xyb.ConstPlaneRow(1, y1);
+      const float* row_in2 = xyb.ConstPlaneRow(1, y2);
       float* mask1x1_out = mask1x1->Row(y);
       auto scalar_pixel1x1 = [&](size_t x) {
         const size_t x2 = x + 1 < xsize ? x + 1 : x;
@@ -460,15 +464,21 @@ struct AdaptiveQuantizationImpl {
         static const float kOffset = 0.01;
         mask1x1_out[x] = kMul / (diff + kOffset);
       };
-      for (size_t x = x_start; x < x_end; ++x) {
+      for (size_t x = x_start_1x1; x < x_end_1x1; ++x) {
         scalar_pixel1x1(x);
       }
     }
 
+    size_t y_start = rect_in.y0() + rect_out.y0() * 8;
+    size_t y_end = y_start + rect_out.ysize() * 8;
+
+    size_t x_start = rect_in.x0() + rect_out.x0() * 8;
+    size_t x_end = x_start + rect_out.xsize() * 8;
+
     if (x_start != 0) x_start -= 4;
-    if (x_end != xyb.xsize()) x_end += 4;
+    if (x_end != xsize) x_end += 4;
     if (y_start != 0) y_start -= 4;
-    if (y_end != xyb.ysize()) y_end += 4;
+    if (y_end != ysize) y_end += 4;
     pre_erosion[thread].ShrinkTo((x_end - x_start) / 4, (y_end - y_start) / 4);
 
     static const float limit = 0.2f;
@@ -476,9 +486,9 @@ struct AdaptiveQuantizationImpl {
       size_t y2 = y + 1 < ysize ? y + 1 : y;
       size_t y1 = y > 0 ? y - 1 : y;
 
-      const float* row_in = xyb.PlaneRow(1, y);
-      const float* row_in1 = xyb.PlaneRow(1, y1);
-      const float* row_in2 = xyb.PlaneRow(1, y2);
+      const float* row_in = xyb.ConstPlaneRow(1, y);
+      const float* row_in1 = xyb.ConstPlaneRow(1, y1);
+      const float* row_in2 = xyb.ConstPlaneRow(1, y2);
       float* JXL_RESTRICT row_out = diff_buffer.Row(thread);
 
       auto scalar_pixel = [&](size_t x) {
@@ -543,26 +553,25 @@ struct AdaptiveQuantizationImpl {
       }
     }
     Rect from_rect(x_start % 8 == 0 ? 0 : 1, y_start % 8 == 0 ? 0 : 1,
-                   rect.xsize() * 2, rect.ysize() * 2);
-    FuzzyErosion(butteraugli_target, from_rect, pre_erosion[thread], rect,
+                   rect_out.xsize() * 2, rect_out.ysize() * 2);
+    FuzzyErosion(butteraugli_target, from_rect, pre_erosion[thread], rect_out,
                  &aq_map);
-    for (size_t y = 0; y < rect.ysize(); ++y) {
-      const float* aq_map_row = rect.ConstRow(aq_map, y);
-      float* mask_row = rect.Row(mask, y);
-      for (size_t x = 0; x < rect.xsize(); ++x) {
+    for (size_t y = 0; y < rect_out.ysize(); ++y) {
+      const float* aq_map_row = rect_out.ConstRow(aq_map, y);
+      float* mask_row = rect_out.Row(mask, y);
+      for (size_t x = 0; x < rect_out.xsize(); ++x) {
         mask_row[x] = ComputeMaskForAcStrategyUse(aq_map_row[x]);
       }
     }
     PerBlockModulations(butteraugli_target, xyb.Plane(0), xyb.Plane(1),
-                        xyb.Plane(2), scale, rect, &aq_map);
+                        xyb.Plane(2), rect_in, scale, rect_out, &aq_map);
   }
   std::vector<ImageF> pre_erosion;
   ImageF aq_map;
   ImageF diff_buffer;
 };
 
-static void Blur1x1Masking(const FrameDimensions& frame_dim, ThreadPool* pool,
-                           ImageF* mask1x1) {
+Status Blur1x1Masking(ThreadPool* pool, ImageF* mask1x1, const Rect& rect) {
   // Blur the mask1x1 to obtain the masking image.
   // Before blurring it contains an image of absolute value of the
   // Laplacian of the intensity channel.
@@ -588,47 +597,46 @@ static void Blur1x1Masking(const FrameDimensions& frame_dim, ThreadPool* pool,
                         {HWY_REP4(normalize_mul * kFilterMask1x1[1])},
                         {HWY_REP4(normalize_mul * kFilterMask1x1[4])},
                         {HWY_REP4(normalize_mul * kFilterMask1x1[3])}};
-  Rect from_rect(0, 0, 8 * frame_dim.xsize_blocks, 8 * frame_dim.ysize_blocks);
-  ImageF temp(mask1x1->xsize(), mask1x1->ysize());
-  Symmetric5(*mask1x1, from_rect, weights, pool, &temp);
-  CopyImageTo(temp, mask1x1);  // TODO: make it a swap
+  JXL_ASSIGN_OR_RETURN(ImageF temp, ImageF::Create(rect.xsize(), rect.ysize()));
+  Symmetric5(*mask1x1, rect, weights, pool, &temp);
+  *mask1x1 = std::move(temp);
+  return true;
 }
 
-ImageF AdaptiveQuantizationMap(const float butteraugli_target,
-                               const Image3F& xyb,
-                               const FrameDimensions& frame_dim, float scale,
-                               ThreadPool* pool, ImageF* mask,
-                               ImageF* mask1x1) {
+StatusOr<ImageF> AdaptiveQuantizationMap(const float butteraugli_target,
+                                         const Image3F& xyb, const Rect& rect,
+                                         float scale, ThreadPool* pool,
+                                         ImageF* mask, ImageF* mask1x1) {
+  JXL_DASSERT(rect.xsize() % kBlockDim == 0);
+  JXL_DASSERT(rect.ysize() % kBlockDim == 0);
   AdaptiveQuantizationImpl impl;
-  impl.Init(xyb);
-  *mask = ImageF(frame_dim.xsize_blocks, frame_dim.ysize_blocks);
-  *mask1x1 = ImageF(8 * frame_dim.xsize_blocks, 8 * frame_dim.ysize_blocks);
+  const size_t xsize_blocks = rect.xsize() / kBlockDim;
+  const size_t ysize_blocks = rect.ysize() / kBlockDim;
+  JXL_ASSIGN_OR_RETURN(impl.aq_map, ImageF::Create(xsize_blocks, ysize_blocks));
+  JXL_ASSIGN_OR_RETURN(*mask, ImageF::Create(xsize_blocks, ysize_blocks));
+  JXL_ASSIGN_OR_RETURN(*mask1x1, ImageF::Create(xyb.xsize(), xyb.ysize()));
   JXL_CHECK(RunOnPool(
       pool, 0,
-      DivCeil(frame_dim.xsize_blocks, kEncTileDimInBlocks) *
-          DivCeil(frame_dim.ysize_blocks, kEncTileDimInBlocks),
+      DivCeil(xsize_blocks, kEncTileDimInBlocks) *
+          DivCeil(ysize_blocks, kEncTileDimInBlocks),
       [&](const size_t num_threads) {
-        impl.PrepareBuffers(num_threads);
-        return true;
+        return !!impl.PrepareBuffers(num_threads);
       },
       [&](const uint32_t tid, const size_t thread) {
-        size_t n_enc_tiles =
-            DivCeil(frame_dim.xsize_blocks, kEncTileDimInBlocks);
+        size_t n_enc_tiles = DivCeil(xsize_blocks, kEncTileDimInBlocks);
         size_t tx = tid % n_enc_tiles;
         size_t ty = tid / n_enc_tiles;
         size_t by0 = ty * kEncTileDimInBlocks;
-        size_t by1 =
-            std::min((ty + 1) * kEncTileDimInBlocks, frame_dim.ysize_blocks);
+        size_t by1 = std::min((ty + 1) * kEncTileDimInBlocks, ysize_blocks);
         size_t bx0 = tx * kEncTileDimInBlocks;
-        size_t bx1 =
-            std::min((tx + 1) * kEncTileDimInBlocks, frame_dim.xsize_blocks);
-        Rect r(bx0, by0, bx1 - bx0, by1 - by0);
-        impl.ComputeTile(butteraugli_target, scale, xyb, r, thread, mask,
-                         mask1x1);
+        size_t bx1 = std::min((tx + 1) * kEncTileDimInBlocks, xsize_blocks);
+        Rect rect_out(bx0, by0, bx1 - bx0, by1 - by0);
+        impl.ComputeTile(butteraugli_target, scale, xyb, rect, rect_out, thread,
+                         mask, mask1x1);
       },
       "AQ DiffPrecompute"));
 
-  Blur1x1Masking(frame_dim, pool, mask1x1);
+  JXL_RETURN_IF_ERROR(Blur1x1Masking(pool, mask1x1, rect));
   return std::move(impl).aq_map;
 }
 
@@ -648,24 +656,28 @@ namespace {
 // If true, prints the quantization maps at each iteration.
 constexpr bool FLAGS_dump_quant_state = false;
 
-void DumpHeatmap(const CompressParams& cparams, const AuxOut* aux_out,
-                 const std::string& label, const ImageF& image,
-                 float good_threshold, float bad_threshold) {
+Status DumpHeatmap(const CompressParams& cparams, const AuxOut* aux_out,
+                   const std::string& label, const ImageF& image,
+                   float good_threshold, float bad_threshold) {
   if (JXL_DEBUG_ADAPTIVE_QUANTIZATION) {
-    Image3F heatmap = CreateHeatMapImage(image, good_threshold, bad_threshold);
+    JXL_ASSIGN_OR_RETURN(
+        Image3F heatmap,
+        CreateHeatMapImage(image, good_threshold, bad_threshold));
     char filename[200];
     snprintf(filename, sizeof(filename), "%s%05d", label.c_str(),
              aux_out->num_butteraugli_iters);
-    DumpImage(cparams, filename, heatmap);
+    JXL_RETURN_IF_ERROR(DumpImage(cparams, filename, heatmap));
   }
+  return true;
 }
 
-void DumpHeatmaps(const CompressParams& cparams, const AuxOut* aux_out,
-                  float ba_target, const ImageF& quant_field,
-                  const ImageF& tile_heatmap, const ImageF& bt_diffmap) {
+Status DumpHeatmaps(const CompressParams& cparams, const AuxOut* aux_out,
+                    float ba_target, const ImageF& quant_field,
+                    const ImageF& tile_heatmap, const ImageF& bt_diffmap) {
   if (JXL_DEBUG_ADAPTIVE_QUANTIZATION) {
-    if (!WantDebugOutput(cparams)) return;
-    ImageF inv_qmap(quant_field.xsize(), quant_field.ysize());
+    if (!WantDebugOutput(cparams)) return true;
+    JXL_ASSIGN_OR_RETURN(ImageF inv_qmap, ImageF::Create(quant_field.xsize(),
+                                                         quant_field.ysize()));
     for (size_t y = 0; y < quant_field.ysize(); ++y) {
       const float* JXL_RESTRICT row_q = quant_field.ConstRow(y);
       float* JXL_RESTRICT row_inv_q = inv_qmap.Row(y);
@@ -673,21 +685,24 @@ void DumpHeatmaps(const CompressParams& cparams, const AuxOut* aux_out,
         row_inv_q[x] = 1.0f / row_q[x];  // never zero
       }
     }
-    DumpHeatmap(cparams, aux_out, "quant_heatmap", inv_qmap, 4.0f * ba_target,
-                6.0f * ba_target);
-    DumpHeatmap(cparams, aux_out, "tile_heatmap", tile_heatmap, ba_target,
-                1.5f * ba_target);
+    JXL_RETURN_IF_ERROR(DumpHeatmap(cparams, aux_out, "quant_heatmap", inv_qmap,
+                                    4.0f * ba_target, 6.0f * ba_target));
+    JXL_RETURN_IF_ERROR(DumpHeatmap(cparams, aux_out, "tile_heatmap",
+                                    tile_heatmap, ba_target, 1.5f * ba_target));
     // matches heat maps produced by the command line tool.
-    DumpHeatmap(cparams, aux_out, "bt_diffmap", bt_diffmap,
-                ButteraugliFuzzyInverse(1.5), ButteraugliFuzzyInverse(0.5));
+    JXL_RETURN_IF_ERROR(DumpHeatmap(cparams, aux_out, "bt_diffmap", bt_diffmap,
+                                    ButteraugliFuzzyInverse(1.5),
+                                    ButteraugliFuzzyInverse(0.5)));
   }
+  return true;
 }
 
-ImageF TileDistMap(const ImageF& distmap, int tile_size, int margin,
-                   const AcStrategyImage& ac_strategy) {
+StatusOr<ImageF> TileDistMap(const ImageF& distmap, int tile_size, int margin,
+                             const AcStrategyImage& ac_strategy) {
   const int tile_xsize = (distmap.xsize() + tile_size - 1) / tile_size;
   const int tile_ysize = (distmap.ysize() + tile_size - 1) / tile_size;
-  ImageF tile_distmap(tile_xsize, tile_ysize);
+  JXL_ASSIGN_OR_RETURN(ImageF tile_distmap,
+                       ImageF::Create(tile_xsize, tile_ysize));
   size_t distmap_stride = tile_distmap.PixelsPerRow();
   for (int tile_y = 0; tile_y < tile_ysize; ++tile_y) {
     AcStrategyRow ac_strategy_row = ac_strategy.ConstRow(tile_y);
@@ -748,13 +763,16 @@ ImageF TileDistMap(const ImageF& distmap, int tile_size, int margin,
   return tile_distmap;
 }
 
-static const float kDcQuantPow = 0.83f;
-static const float kDcQuant = 1.095924047623553f;
-static const float kAcQuant = 0.7381485255235064f;
+const float kDcQuantPow = 0.83f;
+const float kDcQuant = 1.095924047623553f;
+const float kAcQuant = 0.7381485255235064f;
 
 // Computes the decoded image for a given set of compression parameters.
-ImageBundle RoundtripImage(const Image3F& opsin, PassesEncoderState* enc_state,
-                           const JxlCmsInterface& cms, ThreadPool* pool) {
+StatusOr<ImageBundle> RoundtripImage(const FrameHeader& frame_header,
+                                     const Image3F& opsin,
+                                     PassesEncoderState* enc_state,
+                                     const JxlCmsInterface& cms,
+                                     ThreadPool* pool) {
   std::unique_ptr<PassesDecoderState> dec_state =
       jxl::make_unique<PassesDecoderState>();
   JXL_CHECK(dec_state->output_encoding_info.SetFromMetadata(
@@ -767,18 +785,20 @@ ImageBundle RoundtripImage(const Image3F& opsin, PassesEncoderState* enc_state,
   const size_t num_groups = xsize_groups * ysize_groups;
 
   size_t num_special_frames = enc_state->special_frames.size();
-
-  std::unique_ptr<ModularFrameEncoder> modular_frame_encoder =
-      jxl::make_unique<ModularFrameEncoder>(enc_state->shared.frame_header,
-                                            enc_state->cparams);
-  JXL_CHECK(InitializePassesEncoder(opsin, cms, pool, enc_state,
-                                    modular_frame_encoder.get(), nullptr));
-  JXL_CHECK(dec_state->Init());
-  JXL_CHECK(dec_state->InitForAC(pool));
+  size_t num_passes = enc_state->progressive_splitter.GetNumPasses();
+  ModularFrameEncoder modular_frame_encoder(frame_header, enc_state->cparams,
+                                            false);
+  JXL_CHECK(InitializePassesEncoder(frame_header, opsin, Rect(opsin), cms, pool,
+                                    enc_state, &modular_frame_encoder,
+                                    nullptr));
+  JXL_CHECK(dec_state->Init(frame_header));
+  JXL_CHECK(dec_state->InitForAC(num_passes, pool));
 
   ImageBundle decoded(&enc_state->shared.metadata->m);
-  decoded.origin = enc_state->shared.frame_header.frame_origin;
-  decoded.SetFromImage(Image3F(opsin.xsize(), opsin.ysize()),
+  decoded.origin = frame_header.frame_origin;
+  JXL_ASSIGN_OR_RETURN(Image3F tmp,
+                       Image3F::Create(opsin.xsize(), opsin.ysize()));
+  decoded.SetFromImage(std::move(tmp),
                        dec_state->output_encoding_info.color_encoding);
 
   PassesDecoderState::PipelineOptions options;
@@ -787,10 +807,10 @@ ImageBundle RoundtripImage(const Image3F& opsin, PassesEncoderState* enc_state,
   options.render_spotcolors = false;
   options.render_noise = false;
 
-  // Same as dec_state->shared->frame_header.nonserialized_metadata->m
+  // Same as frame_header.nonserialized_metadata->m
   const ImageMetadata& metadata = *decoded.metadata();
 
-  JXL_CHECK(dec_state->PreparePipeline(&decoded, options));
+  JXL_CHECK(dec_state->PreparePipeline(frame_header, &decoded, options));
 
   hwy::AlignedUniquePtr<GroupDecCache[]> group_dec_caches;
   const auto allocate_storage = [&](const size_t num_threads) -> Status {
@@ -800,25 +820,32 @@ ImageBundle RoundtripImage(const Image3F& opsin, PassesEncoderState* enc_state,
     group_dec_caches = hwy::MakeUniqueAlignedArray<GroupDecCache>(num_threads);
     return true;
   };
+  std::atomic<bool> has_error{false};
   const auto process_group = [&](const uint32_t group_index,
                                  const size_t thread) {
-    if (dec_state->shared->frame_header.loop_filter.epf_iters > 0) {
-      ComputeSigma(dec_state->shared->BlockGroupRect(group_index),
+    if (has_error) return;
+    if (frame_header.loop_filter.epf_iters > 0) {
+      ComputeSigma(frame_header.loop_filter,
+                   dec_state->shared->frame_dim.BlockGroupRect(group_index),
                    dec_state.get());
     }
     RenderPipelineInput input =
         dec_state->render_pipeline->GetInputBuffers(group_index, thread);
     JXL_CHECK(DecodeGroupForRoundtrip(
-        enc_state->coeffs, group_index, dec_state.get(),
+        frame_header, enc_state->coeffs, group_index, dec_state.get(),
         &group_dec_caches[thread], thread, input, &decoded, nullptr));
     for (size_t c = 0; c < metadata.num_extra_channels; c++) {
       std::pair<ImageF*, Rect> ri = input.GetBuffer(3 + c);
       FillPlane(0.0f, ri.first, ri.second);
     }
-    input.Done();
+    if (!input.Done()) {
+      has_error = true;
+      return;
+    }
   };
   JXL_CHECK(RunOnPool(pool, 0, num_groups, allocate_storage, process_group,
                       "AQ loop"));
+  if (has_error) return JXL_FAILURE("AQ loop failure");
 
   // Ensure we don't create any new special frames.
   enc_state->special_frames.resize(num_special_frames);
@@ -828,65 +855,41 @@ ImageBundle RoundtripImage(const Image3F& opsin, PassesEncoderState* enc_state,
 
 constexpr int kMaxButteraugliIters = 4;
 
-void FindBestQuantization(const ImageBundle& linear, const Image3F& opsin,
-                          PassesEncoderState* enc_state,
-                          const JxlCmsInterface& cms, ThreadPool* pool,
-                          AuxOut* aux_out) {
+Status FindBestQuantization(const FrameHeader& frame_header,
+                            const Image3F& linear, const Image3F& opsin,
+                            ImageF& quant_field, PassesEncoderState* enc_state,
+                            const JxlCmsInterface& cms, ThreadPool* pool,
+                            AuxOut* aux_out) {
   const CompressParams& cparams = enc_state->cparams;
   if (cparams.resampling > 1 &&
       cparams.original_butteraugli_distance <= 4.0 * cparams.resampling) {
     // For downsampled opsin image, the butteraugli based adaptive quantization
     // loop would only make the size bigger without improving the distance much,
     // so in this case we enable it only for very high butteraugli targets.
-    return;
+    return true;
   }
   Quantizer& quantizer = enc_state->shared.quantizer;
   ImageI& raw_quant_field = enc_state->shared.raw_quant_field;
-  ImageF& quant_field = enc_state->initial_quant_field;
-
-  // TODO(veluca): this should really be rather handled on the
-  // ButteraugliComparator side.
-  struct TemporaryShrink {
-    TemporaryShrink(ImageBundle& bundle, size_t xsize, size_t ysize)
-        : bundle(bundle),
-          orig_xsize(bundle.xsize()),
-          orig_ysize(bundle.ysize()) {
-      bundle.ShrinkTo(xsize, ysize);
-    }
-    TemporaryShrink(const TemporaryShrink&) = delete;
-    TemporaryShrink(TemporaryShrink&&) = delete;
-
-    ~TemporaryShrink() { bundle.ShrinkTo(orig_xsize, orig_ysize); }
-
-    ImageBundle& bundle;
-    size_t orig_xsize;
-    size_t orig_ysize;
-  } t(const_cast<ImageBundle&>(linear),
-      enc_state->shared.frame_header.frame_size.xsize,
-      enc_state->shared.frame_header.frame_size.ysize);
 
   const float butteraugli_target = cparams.butteraugli_distance;
   const float original_butteraugli = cparams.original_butteraugli_distance;
   ButteraugliParams params;
-  params.intensity_target = linear.metadata()->IntensityTarget();
-  // Hack the default intensity target value to be 80.0, the intensity
-  // target of sRGB images and a more reasonable viewing default than
-  // JPEG XL file format's default.
-  if (fabs(params.intensity_target - 255.0f) < 1e-3) {
-    params.intensity_target = 80.0f;
-  }
+  params.intensity_target = 80.f;
   JxlButteraugliComparator comparator(params, cms);
-  JXL_CHECK(comparator.SetReferenceImage(linear));
+  JXL_CHECK(comparator.SetLinearReferenceImage(linear));
   bool lower_is_better =
       (comparator.GoodQualityScore() < comparator.BadQualityScore());
   const float initial_quant_dc = InitialQuantDC(butteraugli_target);
   AdjustQuantField(enc_state->shared.ac_strategy, Rect(quant_field),
                    original_butteraugli, &quant_field);
   ImageF tile_distmap;
-  ImageF initial_quant_field(quant_field.xsize(), quant_field.ysize());
+  JXL_ASSIGN_OR_RETURN(
+      ImageF initial_quant_field,
+      ImageF::Create(quant_field.xsize(), quant_field.ysize()));
   CopyImageTo(quant_field, &initial_quant_field);
 
-  float initial_qf_min, initial_qf_max;
+  float initial_qf_min;
+  float initial_qf_max;
   ImageMinMax(initial_quant_field, &initial_qf_min, &initial_qf_max);
   float initial_qf_ratio = initial_qf_max / initial_qf_min;
   float qf_max_deviation_low = std::sqrt(250 / initial_qf_ratio);
@@ -913,7 +916,9 @@ void FindBestQuantization(const ImageBundle& linear, const Image3F& opsin,
       }
     }
     quantizer.SetQuantField(initial_quant_dc, quant_field, &raw_quant_field);
-    ImageBundle dec_linear = RoundtripImage(opsin, enc_state, cms, pool);
+    JXL_ASSIGN_OR_RETURN(
+        ImageBundle dec_linear,
+        RoundtripImage(frame_header, opsin, enc_state, cms, pool));
     float score;
     ImageF diffmap;
     JXL_CHECK(comparator.CompareWith(dec_linear, &diffmap, &score));
@@ -921,16 +926,19 @@ void FindBestQuantization(const ImageBundle& linear, const Image3F& opsin,
       score = -score;
       ScaleImage(-1.0f, &diffmap);
     }
-    tile_distmap = TileDistMap(diffmap, 8 * cparams.resampling, 0,
-                               enc_state->shared.ac_strategy);
+    JXL_ASSIGN_OR_RETURN(tile_distmap,
+                         TileDistMap(diffmap, 8 * cparams.resampling, 0,
+                                     enc_state->shared.ac_strategy));
     if (JXL_DEBUG_ADAPTIVE_QUANTIZATION && WantDebugOutput(cparams)) {
-      DumpImage(cparams, ("dec" + ToString(i)).c_str(), *dec_linear.color());
-      DumpHeatmaps(cparams, aux_out, butteraugli_target, quant_field,
-                   tile_distmap, diffmap);
+      JXL_RETURN_IF_ERROR(DumpImage(cparams, ("dec" + ToString(i)).c_str(),
+                                    *dec_linear.color()));
+      JXL_RETURN_IF_ERROR(DumpHeatmaps(cparams, aux_out, butteraugli_target,
+                                       quant_field, tile_distmap, diffmap));
     }
     if (aux_out != nullptr) ++aux_out->num_butteraugli_iters;
     if (JXL_DEBUG_ADAPTIVE_QUANTIZATION) {
-      float minval, maxval;
+      float minval;
+      float maxval;
       ImageMinMax(quant_field, &minval, &maxval);
       printf("\nButteraugli iter: %d/%d\n", i, kMaxButteraugliIters);
       printf("Butteraugli distance: %f  (target = %f)\n", score,
@@ -1020,17 +1028,18 @@ void FindBestQuantization(const ImageBundle& linear, const Image3F& opsin,
     }
   }
   quantizer.SetQuantField(initial_quant_dc, quant_field, &raw_quant_field);
+  return true;
 }
 
-void FindBestQuantizationMaxError(const Image3F& opsin,
-                                  PassesEncoderState* enc_state,
-                                  const JxlCmsInterface& cms, ThreadPool* pool,
-                                  AuxOut* aux_out) {
+Status FindBestQuantizationMaxError(const FrameHeader& frame_header,
+                                    const Image3F& opsin, ImageF& quant_field,
+                                    PassesEncoderState* enc_state,
+                                    const JxlCmsInterface& cms,
+                                    ThreadPool* pool, AuxOut* aux_out) {
   // TODO(szabadka): Make this work for non-opsin color spaces.
   const CompressParams& cparams = enc_state->cparams;
   Quantizer& quantizer = enc_state->shared.quantizer;
   ImageI& raw_quant_field = enc_state->shared.raw_quant_field;
-  ImageF& quant_field = enc_state->initial_quant_field;
 
   // TODO(veluca): better choice of this value.
   const float initial_quant_dc =
@@ -1045,11 +1054,15 @@ void FindBestQuantizationMaxError(const Image3F& opsin,
   for (int i = 0; i < kMaxButteraugliIters + 1; ++i) {
     quantizer.SetQuantField(initial_quant_dc, quant_field, &raw_quant_field);
     if (JXL_DEBUG_ADAPTIVE_QUANTIZATION && aux_out) {
-      DumpXybImage(cparams, ("ops" + ToString(i)).c_str(), opsin);
+      JXL_RETURN_IF_ERROR(
+          DumpXybImage(cparams, ("ops" + ToString(i)).c_str(), opsin));
     }
-    ImageBundle decoded = RoundtripImage(opsin, enc_state, cms, pool);
+    JXL_ASSIGN_OR_RETURN(
+        ImageBundle decoded,
+        RoundtripImage(frame_header, opsin, enc_state, cms, pool));
     if (JXL_DEBUG_ADAPTIVE_QUANTIZATION && aux_out) {
-      DumpXybImage(cparams, ("dec" + ToString(i)).c_str(), *decoded.color());
+      JXL_RETURN_IF_ERROR(DumpXybImage(cparams, ("dec" + ToString(i)).c_str(),
+                                       *decoded.color()));
     }
     for (size_t by = 0; by < enc_state->shared.frame_dim.ysize_blocks; by++) {
       AcStrategyRow ac_strategy_row =
@@ -1091,6 +1104,7 @@ void FindBestQuantizationMaxError(const Image3F& opsin,
     }
   }
   quantizer.SetQuantField(initial_quant_dc, quant_field, &raw_quant_field);
+  return true;
 }
 
 }  // namespace
@@ -1160,26 +1174,31 @@ float InitialQuantDC(float butteraugli_target) {
   return std::min(kDcQuant / butteraugli_target_dc, 50.f);
 }
 
-ImageF InitialQuantField(const float butteraugli_target, const Image3F& opsin,
-                         const FrameDimensions& frame_dim, ThreadPool* pool,
-                         float rescale, ImageF* mask, ImageF* mask1x1) {
+StatusOr<ImageF> InitialQuantField(const float butteraugli_target,
+                                   const Image3F& opsin, const Rect& rect,
+                                   ThreadPool* pool, float rescale,
+                                   ImageF* mask, ImageF* mask1x1) {
   const float quant_ac = kAcQuant / butteraugli_target;
   return HWY_DYNAMIC_DISPATCH(AdaptiveQuantizationMap)(
-      butteraugli_target, opsin, frame_dim, quant_ac * rescale, pool, mask,
-      mask1x1);
+      butteraugli_target, opsin, rect, quant_ac * rescale, pool, mask, mask1x1);
 }
 
-void FindBestQuantizer(const ImageBundle* linear, const Image3F& opsin,
-                       PassesEncoderState* enc_state,
-                       const JxlCmsInterface& cms, ThreadPool* pool,
-                       AuxOut* aux_out, double rescale) {
+Status FindBestQuantizer(const FrameHeader& frame_header, const Image3F* linear,
+                         const Image3F& opsin, ImageF& quant_field,
+                         PassesEncoderState* enc_state,
+                         const JxlCmsInterface& cms, ThreadPool* pool,
+                         AuxOut* aux_out, double rescale) {
   const CompressParams& cparams = enc_state->cparams;
   if (cparams.max_error_mode) {
-    FindBestQuantizationMaxError(opsin, enc_state, cms, pool, aux_out);
-  } else if (cparams.speed_tier <= SpeedTier::kKitten) {
+    JXL_RETURN_IF_ERROR(FindBestQuantizationMaxError(
+        frame_header, opsin, quant_field, enc_state, cms, pool, aux_out));
+  } else if (linear && cparams.speed_tier <= SpeedTier::kKitten) {
     // Normal encoding to a butteraugli score.
-    FindBestQuantization(*linear, opsin, enc_state, cms, pool, aux_out);
+    JXL_RETURN_IF_ERROR(FindBestQuantization(frame_header, *linear, opsin,
+                                             quant_field, enc_state, cms, pool,
+                                             aux_out));
   }
+  return true;
 }
 
 }  // namespace jxl
