@@ -10,6 +10,7 @@
 #include <algorithm>
 #include <array>
 #include <cmath>
+#include <cstdint>
 #include <limits>
 #include <numeric>
 #include <type_traits>
@@ -18,23 +19,29 @@
 #include <vector>
 
 #include "lib/jxl/ans_common.h"
-#include "lib/jxl/aux_out.h"
-#include "lib/jxl/aux_out_fwd.h"
 #include "lib/jxl/base/bits.h"
+#include "lib/jxl/base/fast_math-inl.h"
+#include "lib/jxl/base/status.h"
 #include "lib/jxl/dec_ans.h"
+#include "lib/jxl/enc_ans_params.h"
+#include "lib/jxl/enc_aux_out.h"
 #include "lib/jxl/enc_cluster.h"
 #include "lib/jxl/enc_context_map.h"
+#include "lib/jxl/enc_fields.h"
 #include "lib/jxl/enc_huffman.h"
-#include "lib/jxl/fast_math-inl.h"
+#include "lib/jxl/enc_params.h"
 #include "lib/jxl/fields.h"
 
 namespace jxl {
 
 namespace {
 
-bool ans_fuzzer_friendly_ = false;
+#if !JXL_IS_DEBUG_BUILD
+constexpr
+#endif
+    bool ans_fuzzer_friendly_ = false;
 
-static const int kMaxNumSymbolsForSmallCode = 4;
+const int kMaxNumSymbolsForSmallCode = 4;
 
 void ANSBuildInfoTable(const ANSHistBin* counts, const AliasTable::Entry* table,
                        size_t alphabet_size, size_t log_alpha_size,
@@ -96,16 +103,16 @@ float EstimateDataBitsFlat(const ANSHistBin* histogram, size_t len) {
 
 // Static Huffman code for encoding logcounts. The last symbol is used as RLE
 // sequence.
-static const uint8_t kLogCountBitLengths[ANS_LOG_TAB_SIZE + 2] = {
+const uint8_t kLogCountBitLengths[ANS_LOG_TAB_SIZE + 2] = {
     5, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 6, 7, 7,
 };
-static const uint8_t kLogCountSymbols[ANS_LOG_TAB_SIZE + 2] = {
+const uint8_t kLogCountSymbols[ANS_LOG_TAB_SIZE + 2] = {
     17, 11, 15, 3, 9, 7, 4, 2, 5, 6, 0, 33, 1, 65,
 };
 
 // Returns the difference between largest count that can be represented and is
 // smaller than "count" and smallest representable count larger than "count".
-static int SmallestIncrement(uint32_t count, uint32_t shift) {
+int SmallestIncrement(uint32_t count, uint32_t shift) {
   int bits = count == 0 ? -1 : FloorLog2Nonzero(count);
   int drop_bits = bits - GetPopulationCountPrecision(bits, shift);
   return drop_bits < 0 ? 1 : (1 << drop_bits);
@@ -450,7 +457,7 @@ size_t BuildAndStoreANSEncodingData(
             &tmp_writer, 8 * alphabet_size + 8);  // safe upper bound
         BuildAndStoreHuffmanTree(histo.data(), alphabet_size, depths.data(),
                                  bits.data(), &tmp_writer);
-        ReclaimAndCharge(&tmp_writer, &allotment, 0, /*aux_out=*/nullptr);
+        allotment.ReclaimAndCharge(&tmp_writer, 0, /*aux_out=*/nullptr);
         cost = tmp_writer.BitsWritten();
       } else {
         size_t start = writer->BitsWritten();
@@ -470,14 +477,6 @@ size_t BuildAndStoreANSEncodingData(
     return cost;
   }
   JXL_ASSERT(alphabet_size <= ANS_TAB_SIZE);
-  // Ensure we ignore trailing zeros in the histogram.
-  if (alphabet_size != 0) {
-    size_t largest_symbol = 0;
-    for (size_t i = 0; i < alphabet_size; i++) {
-      if (histogram[i] != 0) largest_symbol = i;
-    }
-    alphabet_size = largest_symbol + 1;
-  }
   float cost;
   uint32_t method = ComputeBestMethod(histogram, alphabet_size, &cost,
                                       ans_histogram_strategy);
@@ -494,28 +493,26 @@ size_t BuildAndStoreANSEncodingData(
       counts[0] = ANS_TAB_SIZE;
     }
   }
-  if (method == 0) {
-    counts = CreateFlatHistogram(alphabet_size, ANS_TAB_SIZE);
-    AliasTable::Entry a[ANS_MAX_ALPHABET_SIZE];
-    InitAliasTable(counts, ANS_TAB_SIZE, log_alpha_size, a);
-    ANSBuildInfoTable(counts.data(), a, alphabet_size, log_alpha_size, info);
-    if (writer != nullptr) {
-      EncodeFlatHistogram(alphabet_size, writer);
-    }
-    return cost;
-  }
   int omit_pos = 0;
   uint32_t shift = method - 1;
-  JXL_CHECK(NormalizeCounts(counts.data(), &omit_pos, alphabet_size,
-                            ANS_LOG_TAB_SIZE, shift, &num_symbols, symbols));
+  if (method == 0) {
+    counts = CreateFlatHistogram(alphabet_size, ANS_TAB_SIZE);
+  } else {
+    JXL_CHECK(NormalizeCounts(counts.data(), &omit_pos, alphabet_size,
+                              ANS_LOG_TAB_SIZE, shift, &num_symbols, symbols));
+  }
   AliasTable::Entry a[ANS_MAX_ALPHABET_SIZE];
   InitAliasTable(counts, ANS_TAB_SIZE, log_alpha_size, a);
   ANSBuildInfoTable(counts.data(), a, alphabet_size, log_alpha_size, info);
   if (writer != nullptr) {
-    bool ok = EncodeCounts(counts.data(), alphabet_size, omit_pos, num_symbols,
-                           shift, symbols, writer);
-    (void)ok;
-    JXL_DASSERT(ok);
+    if (method == 0) {
+      EncodeFlatHistogram(alphabet_size, writer);
+    } else {
+      bool ok = EncodeCounts(counts.data(), alphabet_size, omit_pos,
+                             num_symbols, method - 1, symbols, writer);
+      (void)ok;
+      JXL_DASSERT(ok);
+    }
   }
   return cost;
 }
@@ -560,8 +557,9 @@ void ChooseUintConfigs(const HistogramParams& params,
                        std::vector<Histogram>* clustered_histograms,
                        EntropyEncodingData* codes, size_t* log_alpha_size) {
   codes->uint_config.resize(clustered_histograms->size());
-
-  if (params.uint_method == HistogramParams::HybridUintMethod::kNone) return;
+  if (params.uint_method == HistogramParams::HybridUintMethod::kNone) {
+    return;
+  }
   if (params.uint_method == HistogramParams::HybridUintMethod::k000) {
     codes->uint_config.clear();
     codes->uint_config.resize(clustered_histograms->size(),
@@ -575,6 +573,12 @@ void ChooseUintConfigs(const HistogramParams& params,
     return;
   }
 
+  // If the uint config is adaptive, just stick with the default in streaming
+  // mode.
+  if (params.streaming_mode) {
+    return;
+  }
+
   // Brute-force method that tries a few options.
   std::vector<HybridUintConfig> configs;
   if (params.uint_method == HistogramParams::HybridUintMethod::kBest) {
@@ -682,6 +686,23 @@ void ChooseUintConfigs(const HistogramParams& params,
 #endif
 }
 
+Histogram HistogramFromSymbolInfo(
+    const std::vector<ANSEncSymbolInfo>& encoding_info, bool use_prefix_code) {
+  Histogram histo;
+  histo.data_.resize(DivCeil(encoding_info.size(), Histogram::kRounding) *
+                     Histogram::kRounding);
+  histo.total_count_ = 0;
+  for (size_t i = 0; i < encoding_info.size(); ++i) {
+    const ANSEncSymbolInfo& info = encoding_info[i];
+    int count = use_prefix_code
+                    ? (info.depth ? (1u << (PREFIX_MAX_BITS - info.depth)) : 0)
+                    : info.freq_;
+    histo.data_[i] = count;
+    histo.total_count_ += count;
+  }
+  return histo;
+}
+
 class HistogramBuilder {
  public:
   explicit HistogramBuilder(const size_t num_contexts)
@@ -696,21 +717,28 @@ class HistogramBuilder {
   size_t BuildAndStoreEntropyCodes(
       const HistogramParams& params,
       const std::vector<std::vector<Token>>& tokens, EntropyEncodingData* codes,
-      std::vector<uint8_t>* context_map, bool use_prefix_code,
-      BitWriter* writer, size_t layer, AuxOut* aux_out) const {
+      std::vector<uint8_t>* context_map, BitWriter* writer, size_t layer,
+      AuxOut* aux_out) const {
+    const size_t prev_histograms = codes->encoding_info.size();
     size_t cost = 0;
-    codes->encoding_info.clear();
-    std::vector<Histogram> clustered_histograms(histograms_);
-    context_map->resize(histograms_.size());
+    std::vector<Histogram> clustered_histograms;
+    for (size_t i = 0; i < prev_histograms; ++i) {
+      clustered_histograms.push_back(HistogramFromSymbolInfo(
+          codes->encoding_info[i], codes->use_prefix_code));
+    }
+    size_t context_offset = context_map->size();
+    context_map->resize(context_offset + histograms_.size());
     if (histograms_.size() > 1) {
       if (!ans_fuzzer_friendly_) {
         std::vector<uint32_t> histogram_symbols;
         ClusterHistograms(params, histograms_, kClustersLimit,
                           &clustered_histograms, &histogram_symbols);
         for (size_t c = 0; c < histograms_.size(); ++c) {
-          (*context_map)[c] = static_cast<uint8_t>(histogram_symbols[c]);
+          (*context_map)[context_offset + c] =
+              static_cast<uint8_t>(histogram_symbols[c]);
         }
       } else {
+        JXL_ASSERT(codes->encoding_info.empty());
         fill(context_map->begin(), context_map->end(), 0);
         size_t max_symbol = 0;
         for (const Histogram& h : histograms_) {
@@ -727,14 +755,16 @@ class HistogramBuilder {
         EncodeContextMap(*context_map, clustered_histograms.size(), writer,
                          layer, aux_out);
       }
+    } else {
+      JXL_ASSERT(codes->encoding_info.empty());
+      clustered_histograms.push_back(histograms_[0]);
     }
     if (aux_out != nullptr) {
-      for (size_t i = 0; i < clustered_histograms.size(); ++i) {
+      for (size_t i = prev_histograms; i < clustered_histograms.size(); ++i) {
         aux_out->layers[layer].clustered_entropy +=
             clustered_histograms[i].ShannonEntropy();
       }
     }
-    codes->use_prefix_code = use_prefix_code;
     size_t log_alpha_size = codes->lz77.enabled ? 8 : 7;  // Sane default.
     if (ans_fuzzer_friendly_) {
       codes->uint_config.clear();
@@ -744,11 +774,15 @@ class HistogramBuilder {
                         codes, &log_alpha_size);
     }
     if (log_alpha_size < 5) log_alpha_size = 5;
+    if (params.streaming_mode) {
+      // TODO(szabadka) Figure out if we can use lower values here.
+      log_alpha_size = 8;
+    }
     SizeWriter size_writer;  // Used if writer == nullptr to estimate costs.
     cost += 1;
-    if (writer) writer->Write(1, use_prefix_code);
+    if (writer) writer->Write(1, codes->use_prefix_code);
 
-    if (use_prefix_code) {
+    if (codes->use_prefix_code) {
       log_alpha_size = PREFIX_MAX_BITS;
     } else {
       cost += 2;
@@ -756,38 +790,39 @@ class HistogramBuilder {
     if (writer == nullptr) {
       EncodeUintConfigs(codes->uint_config, &size_writer, log_alpha_size);
     } else {
-      if (!use_prefix_code) writer->Write(2, log_alpha_size - 5);
+      if (!codes->use_prefix_code) writer->Write(2, log_alpha_size - 5);
       EncodeUintConfigs(codes->uint_config, writer, log_alpha_size);
     }
-    if (use_prefix_code) {
+    if (codes->use_prefix_code) {
       for (size_t c = 0; c < clustered_histograms.size(); ++c) {
-        size_t num_symbol = 1;
-        for (size_t i = 0; i < clustered_histograms[c].data_.size(); i++) {
-          if (clustered_histograms[c].data_[i]) num_symbol = i + 1;
-        }
+        size_t alphabet_size = clustered_histograms[c].alphabet_size();
         if (writer) {
-          StoreVarLenUint16(num_symbol - 1, writer);
+          StoreVarLenUint16(alphabet_size - 1, writer);
         } else {
-          StoreVarLenUint16(num_symbol - 1, &size_writer);
+          StoreVarLenUint16(alphabet_size - 1, &size_writer);
         }
       }
     }
     cost += size_writer.size;
-    for (size_t c = 0; c < clustered_histograms.size(); ++c) {
-      size_t num_symbol = 1;
-      for (size_t i = 0; i < clustered_histograms[c].data_.size(); i++) {
-        if (clustered_histograms[c].data_[i]) num_symbol = i + 1;
-      }
+    for (size_t c = prev_histograms; c < clustered_histograms.size(); ++c) {
+      size_t alphabet_size = clustered_histograms[c].alphabet_size();
       codes->encoding_info.emplace_back();
-      codes->encoding_info.back().resize(std::max<size_t>(1, num_symbol));
-
-      BitWriter::Allotment allotment(writer, 256 + num_symbol * 24);
+      codes->encoding_info.back().resize(alphabet_size);
+      BitWriter* histo_writer = writer;
+      if (params.streaming_mode) {
+        codes->encoded_histograms.emplace_back();
+        histo_writer = &codes->encoded_histograms.back();
+      }
+      BitWriter::Allotment allotment(histo_writer, 256 + alphabet_size * 24);
       cost += BuildAndStoreANSEncodingData(
           params.ans_histogram_strategy, clustered_histograms[c].data_.data(),
-          num_symbol, log_alpha_size, use_prefix_code,
-          codes->encoding_info.back().data(), writer);
-      allotment.FinishedHistogram(writer);
-      ReclaimAndCharge(writer, &allotment, layer, aux_out);
+          alphabet_size, log_alpha_size, codes->use_prefix_code,
+          codes->encoding_info.back().data(), histo_writer);
+      allotment.FinishedHistogram(histo_writer);
+      allotment.ReclaimAndCharge(histo_writer, layer, aux_out);
+      if (params.streaming_mode) {
+        writer->AppendUnaligned(*histo_writer);
+      }
     }
     return cost;
   }
@@ -1015,9 +1050,9 @@ struct HashChain {
     if (pos + 2 < size_) {
       // TODO(lode): take the MSB's of the uint32_t values into account as well,
       // given that the hash code itself is less than 32 bits.
-      result ^= (uint32_t)(data_[pos + 0] << 0u);
-      result ^= (uint32_t)(data_[pos + 1] << hash_shift_);
-      result ^= (uint32_t)(data_[pos + 2] << (hash_shift_ * 2));
+      result ^= static_cast<uint32_t>(data_[pos + 0] << 0u);
+      result ^= static_cast<uint32_t>(data_[pos + 1] << hash_shift_);
+      result ^= static_cast<uint32_t>(data_[pos + 2] << (hash_shift_ * 2));
     } else {
       // No need to compute hash of last 2 bytes, the length 2 is too short.
       return 0;
@@ -1045,7 +1080,7 @@ struct HashChain {
     uint32_t hashval = GetHash(pos);
     uint32_t wpos = pos & window_mask_;
 
-    val[wpos] = (int)hashval;
+    val[wpos] = static_cast<int>(hashval);
     if (head[hashval] != -1) chain[wpos] = head[hashval];
     head[hashval] = wpos;
 
@@ -1116,7 +1151,10 @@ struct HashChain {
       } else {
         if (hashpos == chain[hashpos]) break;
         hashpos = chain[hashpos];
-        if (val[hashpos] != (int)hashval) break;  // outdated hash value
+        if (val[hashpos] != static_cast<int>(hashval)) {
+          // outdated hash value
+          break;
+        }
       }
     }
   }
@@ -1448,7 +1486,9 @@ void ApplyLZ77_Optimal(const HistogramParams& params, size_t num_contexts,
 void ApplyLZ77(const HistogramParams& params, size_t num_contexts,
                const std::vector<std::vector<Token>>& tokens, LZ77Params& lz77,
                std::vector<std::vector<Token>>& tokens_lz77) {
-  lz77.enabled = false;
+  if (params.initialize_global_state) {
+    lz77.enabled = false;
+  }
   if (params.force_huffman) {
     lz77.min_symbol = std::min(PREFIX_MAX_ALPHABET_SIZE - 32, 512);
   } else {
@@ -1463,11 +1503,43 @@ void ApplyLZ77(const HistogramParams& params, size_t num_contexts,
   } else if (params.lz77_method == HistogramParams::LZ77Method::kOptimal) {
     ApplyLZ77_Optimal(params, num_contexts, tokens, lz77, tokens_lz77);
   } else {
-    JXL_ABORT("Not implemented");
+    JXL_UNREACHABLE("Not implemented");
   }
 }
 }  // namespace
 
+void EncodeHistograms(const std::vector<uint8_t>& context_map,
+                      const EntropyEncodingData& codes, BitWriter* writer,
+                      size_t layer, AuxOut* aux_out) {
+  BitWriter::Allotment allotment(writer, 128 + kClustersLimit * 136);
+  JXL_CHECK(Bundle::Write(codes.lz77, writer, layer, aux_out));
+  if (codes.lz77.enabled) {
+    EncodeUintConfig(codes.lz77.length_uint_config, writer,
+                     /*log_alpha_size=*/8);
+  }
+  EncodeContextMap(context_map, codes.encoding_info.size(), writer, layer,
+                   aux_out);
+  writer->Write(1, codes.use_prefix_code);
+  size_t log_alpha_size = 8;
+  if (codes.use_prefix_code) {
+    log_alpha_size = PREFIX_MAX_BITS;
+  } else {
+    log_alpha_size = 8;  // streaming_mode
+    writer->Write(2, log_alpha_size - 5);
+  }
+  EncodeUintConfigs(codes.uint_config, writer, log_alpha_size);
+  if (codes.use_prefix_code) {
+    for (const auto& info : codes.encoding_info) {
+      StoreVarLenUint16(info.size() - 1, writer);
+    }
+  }
+  for (const auto& histo_writer : codes.encoded_histograms) {
+    writer->AppendUnaligned(histo_writer);
+  }
+  allotment.FinishedHistogram(writer);
+  allotment.ReclaimAndCharge(writer, layer, aux_out);
+}
+
 size_t BuildAndEncodeHistograms(const HistogramParams& params,
                                 size_t num_contexts,
                                 std::vector<std::vector<Token>>& tokens,
@@ -1553,28 +1625,59 @@ size_t BuildAndEncodeHistograms(const HistogramParams& params,
     }
   }
 
-  bool use_prefix_code =
-      params.force_huffman || total_tokens < 100 ||
-      params.clustering == HistogramParams::ClusteringType::kFastest ||
-      ans_fuzzer_friendly_;
-  if (!use_prefix_code) {
-    bool all_singleton = true;
-    for (size_t i = 0; i < num_contexts; i++) {
-      if (builder.Histo(i).ShannonEntropy() >= 1e-5) {
-        all_singleton = false;
+  if (params.add_missing_symbols) {
+    for (size_t c = 0; c < num_contexts; ++c) {
+      for (int symbol = 0; symbol < ANS_MAX_ALPHABET_SIZE; ++symbol) {
+        builder.VisitSymbol(symbol, c);
       }
     }
-    if (all_singleton) {
-      use_prefix_code = true;
+  }
+
+  if (params.initialize_global_state) {
+    bool use_prefix_code =
+        params.force_huffman || total_tokens < 100 ||
+        params.clustering == HistogramParams::ClusteringType::kFastest ||
+        ans_fuzzer_friendly_;
+    if (!use_prefix_code) {
+      bool all_singleton = true;
+      for (size_t i = 0; i < num_contexts; i++) {
+        if (builder.Histo(i).ShannonEntropy() >= 1e-5) {
+          all_singleton = false;
+        }
+      }
+      if (all_singleton) {
+        use_prefix_code = true;
+      }
     }
+    codes->use_prefix_code = use_prefix_code;
+  }
+
+  if (params.add_fixed_histograms) {
+    // TODO(szabadka) Add more fixed histograms.
+    // TODO(szabadka) Reduce alphabet size by choosing a non-default
+    // uint_config.
+    const size_t alphabet_size = ANS_MAX_ALPHABET_SIZE;
+    const size_t log_alpha_size = 8;
+    JXL_ASSERT(alphabet_size == 1u << log_alpha_size);
+    std::vector<int32_t> counts =
+        CreateFlatHistogram(alphabet_size, ANS_TAB_SIZE);
+    codes->encoding_info.emplace_back();
+    codes->encoding_info.back().resize(alphabet_size);
+    codes->encoded_histograms.emplace_back();
+    BitWriter* histo_writer = &codes->encoded_histograms.back();
+    BitWriter::Allotment allotment(histo_writer, 256 + alphabet_size * 24);
+    BuildAndStoreANSEncodingData(params.ans_histogram_strategy, counts.data(),
+                                 alphabet_size, log_alpha_size,
+                                 codes->use_prefix_code,
+                                 &codes->encoding_info.back()[0], histo_writer);
+    allotment.ReclaimAndCharge(histo_writer, 0, nullptr);
   }
 
   // Encode histograms.
-  total_bits += builder.BuildAndStoreEntropyCodes(params, tokens, codes,
-                                                  context_map, use_prefix_code,
-                                                  writer, layer, aux_out);
+  total_bits += builder.BuildAndStoreEntropyCodes(
+      params, tokens, codes, context_map, writer, layer, aux_out);
   allotment.FinishedHistogram(writer);
-  ReclaimAndCharge(writer, &allotment, layer, aux_out);
+  allotment.ReclaimAndCharge(writer, layer, aux_out);
 
   if (aux_out != nullptr) {
     aux_out->layers[layer].num_clustered_histograms +=
@@ -1585,13 +1688,14 @@ size_t BuildAndEncodeHistograms(const HistogramParams& params,
 
 size_t WriteTokens(const std::vector<Token>& tokens,
                    const EntropyEncodingData& codes,
-                   const std::vector<uint8_t>& context_map, BitWriter* writer) {
+                   const std::vector<uint8_t>& context_map,
+                   size_t context_offset, BitWriter* writer) {
   size_t num_extra_bits = 0;
   if (codes.use_prefix_code) {
     for (size_t i = 0; i < tokens.size(); i++) {
       uint32_t tok, nbits, bits;
       const Token& token = tokens[i];
-      size_t histo = context_map[token.context];
+      size_t histo = context_map[context_offset + token.context];
       (token.is_lz77_length ? codes.lz77.length_uint_config
                             : codes.uint_config[histo])
           .Encode(token.value, &tok, &nbits, &bits);
@@ -1632,13 +1736,14 @@ size_t WriteTokens(const std::vector<Token>& tokens,
   if (codes.lz77.enabled || context_map.size() > 1) {
     for (int i = end - 1; i >= 0; --i) {
       const Token token = tokens[i];
-      const uint8_t histo = context_map[token.context];
+      const uint8_t histo = context_map[context_offset + token.context];
       uint32_t tok, nbits, bits;
       (token.is_lz77_length ? codes.lz77.length_uint_config
                             : codes.uint_config[histo])
           .Encode(tokens[i].value, &tok, &nbits, &bits);
       tok += token.is_lz77_length ? codes.lz77.min_symbol : 0;
       const ANSEncSymbolInfo& info = codes.encoding_info[histo][tok];
+      JXL_DASSERT(info.freq_ > 0);
       // Extra bits first as this is reversed.
       addbits(bits, nbits);
       num_extra_bits += nbits;
@@ -1670,11 +1775,13 @@ size_t WriteTokens(const std::vector<Token>& tokens,
 
 void WriteTokens(const std::vector<Token>& tokens,
                  const EntropyEncodingData& codes,
-                 const std::vector<uint8_t>& context_map, BitWriter* writer,
-                 size_t layer, AuxOut* aux_out) {
-  BitWriter::Allotment allotment(writer, 32 * tokens.size() + 32 * 1024 * 4);
-  size_t num_extra_bits = WriteTokens(tokens, codes, context_map, writer);
-  ReclaimAndCharge(writer, &allotment, layer, aux_out);
+                 const std::vector<uint8_t>& context_map, size_t context_offset,
+                 BitWriter* writer, size_t layer, AuxOut* aux_out) {
+  // Theoretically, we could have 15 prefix code bits + 31 extra bits.
+  BitWriter::Allotment allotment(writer, 46 * tokens.size() + 32 * 1024 * 4);
+  size_t num_extra_bits =
+      WriteTokens(tokens, codes, context_map, context_offset, writer);
+  allotment.ReclaimAndCharge(writer, layer, aux_out);
   if (aux_out != nullptr) {
     aux_out->layers[layer].extra_bits += num_extra_bits;
   }
@@ -1685,4 +1792,51 @@ void SetANSFuzzerFriendly(bool ans_fuzzer_friendly) {
   ans_fuzzer_friendly_ = ans_fuzzer_friendly;
 #endif
 }
+
+HistogramParams HistogramParams::ForModular(
+    const CompressParams& cparams,
+    const std::vector<uint8_t>& extra_dc_precision, bool streaming_mode) {
+  HistogramParams params;
+  params.streaming_mode = streaming_mode;
+  if (cparams.speed_tier > SpeedTier::kKitten) {
+    params.clustering = HistogramParams::ClusteringType::kFast;
+    params.ans_histogram_strategy =
+        cparams.speed_tier > SpeedTier::kThunder
+            ? HistogramParams::ANSHistogramStrategy::kFast
+            : HistogramParams::ANSHistogramStrategy::kApproximate;
+    params.lz77_method =
+        cparams.decoding_speed_tier >= 3 && cparams.modular_mode
+            ? (cparams.speed_tier >= SpeedTier::kFalcon
+                   ? HistogramParams::LZ77Method::kRLE
+                   : HistogramParams::LZ77Method::kLZ77)
+            : HistogramParams::LZ77Method::kNone;
+    // Near-lossless DC, as well as modular mode, require choosing hybrid uint
+    // more carefully.
+    if ((!extra_dc_precision.empty() && extra_dc_precision[0] != 0) ||
+        (cparams.modular_mode && cparams.speed_tier < SpeedTier::kCheetah)) {
+      params.uint_method = HistogramParams::HybridUintMethod::kFast;
+    } else {
+      params.uint_method = HistogramParams::HybridUintMethod::kNone;
+    }
+  } else if (cparams.speed_tier <= SpeedTier::kTortoise) {
+    params.lz77_method = HistogramParams::LZ77Method::kOptimal;
+  } else {
+    params.lz77_method = HistogramParams::LZ77Method::kLZ77;
+  }
+  if (cparams.decoding_speed_tier >= 1) {
+    params.max_histograms = 12;
+  }
+  if (cparams.decoding_speed_tier >= 1 && cparams.responsive) {
+    params.lz77_method = cparams.speed_tier >= SpeedTier::kCheetah
+                             ? HistogramParams::LZ77Method::kRLE
+                         : cparams.speed_tier >= SpeedTier::kKitten
+                             ? HistogramParams::LZ77Method::kLZ77
+                             : HistogramParams::LZ77Method::kOptimal;
+  }
+  if (cparams.decoding_speed_tier >= 2 && cparams.responsive) {
+    params.uint_method = HistogramParams::HybridUintMethod::k000;
+    params.force_huffman = true;
+  }
+  return params;
+}
 }  // namespace jxl
