@@ -5,31 +5,58 @@
 
 #include "lib/jxl/enc_heuristics.h"
 
+#include <jxl/cms_interface.h>
 #include <stddef.h>
 #include <stdint.h>
 
 #include <algorithm>
+#include <cstdlib>
+#include <limits>
+#include <memory>
 #include <numeric>
 #include <string>
+#include <utility>
+#include <vector>
 
+#include "lib/jxl/ac_context.h"
+#include "lib/jxl/ac_strategy.h"
+#include "lib/jxl/base/common.h"
+#include "lib/jxl/base/compiler_specific.h"
+#include "lib/jxl/base/data_parallel.h"
+#include "lib/jxl/base/override.h"
+#include "lib/jxl/base/status.h"
+#include "lib/jxl/butteraugli/butteraugli.h"
+#include "lib/jxl/chroma_from_luma.h"
+#include "lib/jxl/coeff_order.h"
+#include "lib/jxl/coeff_order_fwd.h"
+#include "lib/jxl/dec_xyb.h"
 #include "lib/jxl/enc_ac_strategy.h"
 #include "lib/jxl/enc_adaptive_quantization.h"
 #include "lib/jxl/enc_ar_control_field.h"
 #include "lib/jxl/enc_cache.h"
 #include "lib/jxl/enc_chroma_from_luma.h"
+#include "lib/jxl/enc_gaborish.h"
 #include "lib/jxl/enc_modular.h"
 #include "lib/jxl/enc_noise.h"
+#include "lib/jxl/enc_params.h"
 #include "lib/jxl/enc_patch_dictionary.h"
-#include "lib/jxl/enc_photon_noise.h"
 #include "lib/jxl/enc_quant_weights.h"
 #include "lib/jxl/enc_splines.h"
-#include "lib/jxl/enc_xyb.h"
-#include "lib/jxl/gaborish.h"
+#include "lib/jxl/frame_dimensions.h"
+#include "lib/jxl/frame_header.h"
+#include "lib/jxl/image.h"
+#include "lib/jxl/image_ops.h"
+#include "lib/jxl/passes_state.h"
+#include "lib/jxl/quant_weights.h"
 
 namespace jxl {
-namespace {
-void FindBestBlockEntropyModel(PassesEncoderState& enc_state) {
-  if (enc_state.cparams.decoding_speed_tier >= 1) {
+
+struct AuxOut;
+
+void FindBestBlockEntropyModel(const CompressParams& cparams, const ImageI& rqf,
+                               const AcStrategyImage& ac_strategy,
+                               BlockCtxMap* block_ctx_map) {
+  if (cparams.decoding_speed_tier >= 1) {
     static constexpr uint8_t kSimpleCtxMap[] = {
         // Cluster all blocks together
         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  //
@@ -40,20 +67,18 @@ void FindBestBlockEntropyModel(PassesEncoderState& enc_state) {
         3 * kNumOrders == sizeof(kSimpleCtxMap) / sizeof *kSimpleCtxMap,
         "Update simple context map");
 
-    auto bcm = enc_state.shared.block_ctx_map;
+    auto bcm = *block_ctx_map;
     bcm.ctx_map.assign(std::begin(kSimpleCtxMap), std::end(kSimpleCtxMap));
     bcm.num_ctxs = 2;
     bcm.num_dc_ctxs = 1;
     return;
   }
-  if (enc_state.cparams.speed_tier >= SpeedTier::kFalcon) {
+  if (cparams.speed_tier >= SpeedTier::kFalcon) {
     return;
   }
-  const ImageI& rqf = enc_state.shared.raw_quant_field;
   // No need to change context modeling for small images.
   size_t tot = rqf.xsize() * rqf.ysize();
-  size_t size_for_ctx_model =
-      (1 << 10) * enc_state.cparams.butteraugli_distance;
+  size_t size_for_ctx_model = (1 << 10) * cparams.butteraugli_distance;
   if (tot < size_for_ctx_model) return;
 
   struct OccCounters {
@@ -77,14 +102,13 @@ void FindBestBlockEntropyModel(PassesEncoderState& enc_state) {
     size_t ord_counts[kNumOrders] = {};
   };
   // The OccCounters struct is too big to allocate on the stack.
-  std::unique_ptr<OccCounters> counters(
-      new OccCounters(rqf, enc_state.shared.ac_strategy));
+  std::unique_ptr<OccCounters> counters(new OccCounters(rqf, ac_strategy));
 
   // Splitting the context model according to the quantization field seems to
   // mostly benefit only large images.
-  size_t size_for_qf_split = (1 << 13) * enc_state.cparams.butteraugli_distance;
+  size_t size_for_qf_split = (1 << 13) * cparams.butteraugli_distance;
   size_t num_qf_segments = tot < size_for_qf_split ? 1 : 2;
-  std::vector<uint32_t>& qft = enc_state.shared.block_ctx_map.qf_thresholds;
+  std::vector<uint32_t>& qft = block_ctx_map->qf_thresholds;
   qft.clear();
   // Divide the quant field in up to num_qf_segments segments.
   size_t cumsum = 0;
@@ -125,8 +149,10 @@ void FindBestBlockEntropyModel(PassesEncoderState& enc_state) {
   std::vector<uint8_t> remap((qft.size() + 1) * kNumOrders);
   std::iota(remap.begin(), remap.end(), 0);
   std::vector<uint8_t> clusters(remap);
-  size_t nb_clusters = Clamp1((int)(tot / size_for_ctx_model / 2), 2, 9);
-  size_t nb_clusters_chroma = Clamp1((int)(tot / size_for_ctx_model / 3), 1, 5);
+  size_t nb_clusters =
+      Clamp1(static_cast<int>(tot / size_for_ctx_model / 2), 2, 9);
+  size_t nb_clusters_chroma =
+      Clamp1(static_cast<int>(tot / size_for_ctx_model / 3), 1, 5);
   // This is O(n^2 log n), but n is small.
   while (clusters.size() > nb_clusters) {
     std::sort(clusters.begin(), clusters.end(),
@@ -151,25 +177,24 @@ void FindBestBlockEntropyModel(PassesEncoderState& enc_state) {
     remap[i] = remap_remap[remap[i]];
   }
   // Write the block context map.
-  auto& ctx_map = enc_state.shared.block_ctx_map.ctx_map;
+  auto& ctx_map = block_ctx_map->ctx_map;
   ctx_map = remap;
   ctx_map.resize(remap.size() * 3);
   // for chroma, only use up to nb_clusters_chroma separate block contexts
   // (those for the biggest clusters)
   for (size_t i = remap.size(); i < remap.size() * 3; i++) {
-    ctx_map[i] = num + Clamp1((int)remap[i % remap.size()], 0,
-                              (int)nb_clusters_chroma - 1);
+    ctx_map[i] = num + Clamp1(static_cast<int>(remap[i % remap.size()]), 0,
+                              static_cast<int>(nb_clusters_chroma) - 1);
   }
-  enc_state.shared.block_ctx_map.num_ctxs =
+  block_ctx_map->num_ctxs =
       *std::max_element(ctx_map.begin(), ctx_map.end()) + 1;
 }
 
-}  // namespace
+namespace {
 
-void FindBestDequantMatrices(const CompressParams& cparams,
-                             const Image3F& opsin,
-                             ModularFrameEncoder* modular_frame_encoder,
-                             DequantMatrices* dequant_matrices) {
+Status FindBestDequantMatrices(const CompressParams& cparams,
+                               ModularFrameEncoder* modular_frame_encoder,
+                               DequantMatrices* dequant_matrices) {
   // TODO(veluca): quant matrices for no-gaborish.
   // TODO(veluca): heuristics for in-bitstream quant tables.
   *dequant_matrices = DequantMatrices();
@@ -181,25 +206,16 @@ void FindBestDequantMatrices(const CompressParams& cparams,
     DctQuantWeightParams dct_params(weights);
     std::vector<QuantEncoding> encodings(DequantMatrices::kNum,
                                          QuantEncoding::DCT(dct_params));
-    DequantMatricesSetCustom(dequant_matrices, encodings,
-                             modular_frame_encoder);
+    JXL_RETURN_IF_ERROR(DequantMatricesSetCustom(dequant_matrices, encodings,
+                                                 modular_frame_encoder));
     float dc_weights[3] = {1.0f / cparams.max_error[0],
                            1.0f / cparams.max_error[1],
                            1.0f / cparams.max_error[2]};
     DequantMatricesSetCustomDC(dequant_matrices, dc_weights);
   }
+  return true;
 }
 
-bool DefaultEncoderHeuristics::HandlesColorConversion(
-    const CompressParams& cparams, const ImageBundle& ib) {
-  return cparams.noise != Override::kOn && cparams.patches != Override::kOn &&
-         cparams.speed_tier >= SpeedTier::kWombat && cparams.resampling == 1 &&
-         cparams.color_transform == ColorTransform::kXYB &&
-         !cparams.modular_mode && !ib.HasAlpha();
-}
-
-namespace {
-
 void StoreMin2(const float v, float& min1, float& min2) {
   if (v < min2) {
     if (v < min1) {
@@ -213,9 +229,9 @@ void StoreMin2(const float v, float& min1, float& min2) {
 
 void CreateMask(const ImageF& image, ImageF& mask) {
   for (size_t y = 0; y < image.ysize(); y++) {
-    auto* row_n = y > 0 ? image.Row(y - 1) : image.Row(y);
-    auto* row_in = image.Row(y);
-    auto* row_s = y + 1 < image.ysize() ? image.Row(y + 1) : image.Row(y);
+    const auto* row_n = y > 0 ? image.Row(y - 1) : image.Row(y);
+    const auto* row_in = image.Row(y);
+    const auto* row_s = y + 1 < image.ysize() ? image.Row(y + 1) : image.Row(y);
     auto* row_out = mask.Row(y);
     for (size_t x = 0; x < image.xsize(); x++) {
       // Center, west, east, north, south values and their absolute difference
@@ -245,7 +261,7 @@ void CreateMask(const ImageF& image, ImageF& mask) {
 // by the decoder. Ringing is slightly reduced by clamping the values of the
 // resulting pixels within certain bounds of a small region in the original
 // image.
-void DownsampleImage2_Sharper(const ImageF& input, ImageF* output) {
+Status DownsampleImage2_Sharper(const ImageF& input, ImageF* output) {
   const int64_t kernelx = 12;
   const int64_t kernely = 12;
 
@@ -302,10 +318,12 @@ void DownsampleImage2_Sharper(const ImageF& input, ImageF* output) {
   int64_t xsize = input.xsize();
   int64_t ysize = input.ysize();
 
-  ImageF box_downsample = CopyImage(input);
-  DownsampleImage(&box_downsample, 2);
+  JXL_ASSIGN_OR_RETURN(ImageF box_downsample, ImageF::Create(xsize, ysize));
+  CopyImageTo(input, &box_downsample);
+  JXL_ASSIGN_OR_RETURN(box_downsample, DownsampleImage(box_downsample, 2));
 
-  ImageF mask(box_downsample.xsize(), box_downsample.ysize());
+  JXL_ASSIGN_OR_RETURN(ImageF mask, ImageF::Create(box_downsample.xsize(),
+                                                   box_downsample.ysize()));
   CreateMask(box_downsample, mask);
 
   for (size_t y = 0; y < output->ysize(); y++) {
@@ -365,46 +383,54 @@ void DownsampleImage2_Sharper(const ImageF& input, ImageF* output) {
       }
     }
   }
+  return true;
 }
 
-void DownsampleImage2_Sharper(Image3F* opsin) {
+}  // namespace
+
+Status DownsampleImage2_Sharper(Image3F* opsin) {
   // Allocate extra space to avoid a reallocation when padding.
-  Image3F downsampled(DivCeil(opsin->xsize(), 2) + kBlockDim,
-                      DivCeil(opsin->ysize(), 2) + kBlockDim);
+  JXL_ASSIGN_OR_RETURN(Image3F downsampled,
+                       Image3F::Create(DivCeil(opsin->xsize(), 2) + kBlockDim,
+                                       DivCeil(opsin->ysize(), 2) + kBlockDim));
   downsampled.ShrinkTo(downsampled.xsize() - kBlockDim,
                        downsampled.ysize() - kBlockDim);
 
   for (size_t c = 0; c < 3; c++) {
-    DownsampleImage2_Sharper(opsin->Plane(c), &downsampled.Plane(c));
+    JXL_RETURN_IF_ERROR(
+        DownsampleImage2_Sharper(opsin->Plane(c), &downsampled.Plane(c)));
   }
   *opsin = std::move(downsampled);
+  return true;
 }
 
-// The default upsampling kernels used by Upsampler in the decoder.
-static const constexpr int64_t kSize = 5;
+namespace {
 
-static const float kernel00[25] = {
+// The default upsampling kernels used by Upsampler in the decoder.
+const constexpr int64_t kSize = 5;
+
+const float kernel00[25] = {
     -0.01716200f, -0.03452303f, -0.04022174f, -0.02921014f, -0.00624645f,
     -0.03452303f, 0.14111091f,  0.28896755f,  0.00278718f,  -0.01610267f,
     -0.04022174f, 0.28896755f,  0.56661550f,  0.03777607f,  -0.01986694f,
     -0.02921014f, 0.00278718f,  0.03777607f,  -0.03144731f, -0.01185068f,
     -0.00624645f, -0.01610267f, -0.01986694f, -0.01185068f, -0.00213539f,
 };
-static const float kernel01[25] = {
+const float kernel01[25] = {
     -0.00624645f, -0.01610267f, -0.01986694f, -0.01185068f, -0.00213539f,
     -0.02921014f, 0.00278718f,  0.03777607f,  -0.03144731f, -0.01185068f,
     -0.04022174f, 0.28896755f,  0.56661550f,  0.03777607f,  -0.01986694f,
     -0.03452303f, 0.14111091f,  0.28896755f,  0.00278718f,  -0.01610267f,
     -0.01716200f, -0.03452303f, -0.04022174f, -0.02921014f, -0.00624645f,
 };
-static const float kernel10[25] = {
+const float kernel10[25] = {
     -0.00624645f, -0.02921014f, -0.04022174f, -0.03452303f, -0.01716200f,
     -0.01610267f, 0.00278718f,  0.28896755f,  0.14111091f,  -0.03452303f,
     -0.01986694f, 0.03777607f,  0.56661550f,  0.28896755f,  -0.04022174f,
     -0.01185068f, -0.03144731f, 0.03777607f,  0.00278718f,  -0.02921014f,
     -0.00213539f, -0.01185068f, -0.01986694f, -0.01610267f, -0.00624645f,
 };
-static const float kernel11[25] = {
+const float kernel11[25] = {
     -0.00213539f, -0.01185068f, -0.01986694f, -0.01610267f, -0.00624645f,
     -0.01185068f, -0.03144731f, 0.03777607f,  0.00278718f,  -0.02921014f,
     -0.01986694f, 0.03777607f,  0.56661550f,  0.28896755f,  -0.04022174f,
@@ -417,14 +443,14 @@ static const float kernel11[25] = {
 // TODO(lode): use Upsampler instead. However, it requires pre-initialization
 // and padding on the left side of the image which requires refactoring the
 // other code using this.
-static void UpsampleImage(const ImageF& input, ImageF* output) {
+void UpsampleImage(const ImageF& input, ImageF* output) {
   int64_t xsize = input.xsize();
   int64_t ysize = input.ysize();
   int64_t xsize2 = output->xsize();
   int64_t ysize2 = output->ysize();
   for (int64_t y = 0; y < ysize2; y++) {
     for (int64_t x = 0; x < xsize2; x++) {
-      auto kernel = kernel00;
+      const auto* kernel = kernel00;
       if ((x & 1) && (y & 1)) {
         kernel = kernel11;
       } else if (x & 1) {
@@ -474,7 +500,7 @@ static void UpsampleImage(const ImageF& input, ImageF* output) {
 // Returns the derivative of Upsampler, with respect to input pixel x2, y2, to
 // output pixel x, y (ignoring the clamping).
 float UpsamplerDeriv(int64_t x2, int64_t y2, int64_t x, int64_t y) {
-  auto kernel = kernel00;
+  const auto* kernel = kernel00;
   if ((x & 1) && (y & 1)) {
     kernel = kernel11;
   } else if (x & 1) {
@@ -579,11 +605,9 @@ void ReduceRinging(const ImageF& initial, const ImageF& mask, ImageF& down) {
       float max = initial.Row(y)[x];
       for (int64_t yi = -1; yi < 2; yi++) {
         for (int64_t xi = -1; xi < 2; xi++) {
-          int64_t x2 = (int64_t)x + xi;
-          int64_t y2 = (int64_t)y + yi;
-          if (x2 < 0 || y2 < 0 || x2 >= (int64_t)xsize2 ||
-              y2 >= (int64_t)ysize2)
-            continue;
+          int64_t x2 = static_cast<int64_t>(x) + xi;
+          int64_t y2 = static_cast<int64_t>(y) + yi;
+          if (x2 < 0 || y2 < 0 || x2 >= xsize2 || y2 >= ysize2) continue;
           min = std::min<float>(min, initial.Row(y2)[x2]);
           max = std::max<float>(max, initial.Row(y2)[x2]);
         }
@@ -607,30 +631,35 @@ void ReduceRinging(const ImageF& initial, const ImageF& mask, ImageF& down) {
 }
 
 // TODO(lode): move this to a separate file enc_downsample.cc
-void DownsampleImage2_Iterative(const ImageF& orig, ImageF* output) {
+Status DownsampleImage2_Iterative(const ImageF& orig, ImageF* output) {
   int64_t xsize = orig.xsize();
   int64_t ysize = orig.ysize();
   int64_t xsize2 = DivCeil(orig.xsize(), 2);
   int64_t ysize2 = DivCeil(orig.ysize(), 2);
 
-  ImageF box_downsample = CopyImage(orig);
-  DownsampleImage(&box_downsample, 2);
-  ImageF mask(box_downsample.xsize(), box_downsample.ysize());
+  JXL_ASSIGN_OR_RETURN(ImageF box_downsample, ImageF::Create(xsize, ysize));
+  CopyImageTo(orig, &box_downsample);
+  JXL_ASSIGN_OR_RETURN(box_downsample, DownsampleImage(box_downsample, 2));
+  JXL_ASSIGN_OR_RETURN(ImageF mask, ImageF::Create(box_downsample.xsize(),
+                                                   box_downsample.ysize()));
   CreateMask(box_downsample, mask);
 
   output->ShrinkTo(xsize2, ysize2);
 
   // Initial result image using the sharper downsampling.
   // Allocate extra space to avoid a reallocation when padding.
-  ImageF initial(DivCeil(orig.xsize(), 2) + kBlockDim,
-                 DivCeil(orig.ysize(), 2) + kBlockDim);
+  JXL_ASSIGN_OR_RETURN(ImageF initial,
+                       ImageF::Create(DivCeil(orig.xsize(), 2) + kBlockDim,
+                                      DivCeil(orig.ysize(), 2) + kBlockDim));
   initial.ShrinkTo(initial.xsize() - kBlockDim, initial.ysize() - kBlockDim);
-  DownsampleImage2_Sharper(orig, &initial);
+  JXL_RETURN_IF_ERROR(DownsampleImage2_Sharper(orig, &initial));
 
-  ImageF down = CopyImage(initial);
-  ImageF up(xsize, ysize);
-  ImageF corr(xsize, ysize);
-  ImageF corr2(xsize2, ysize2);
+  JXL_ASSIGN_OR_RETURN(ImageF down,
+                       ImageF::Create(initial.xsize(), initial.ysize()));
+  CopyImageTo(initial, &down);
+  JXL_ASSIGN_OR_RETURN(ImageF up, ImageF::Create(xsize, ysize));
+  JXL_ASSIGN_OR_RETURN(ImageF corr, ImageF::Create(xsize, ysize));
+  JXL_ASSIGN_OR_RETURN(ImageF corr2, ImageF::Create(xsize2, ysize2));
 
   // In the weights map, relatively higher values will allow less ringing but
   // also less sharpness. With all constant values, it optimizes equally
@@ -639,25 +668,25 @@ void DownsampleImage2_Iterative(const ImageF& orig, ImageF* output) {
   // TODO(lode): Make use of the weights field for anti-ringing and clamping,
   // the values are all set to 1 for now, but it is intended to be used for
   // reducing ringing based on the mask, and taking clamping into account.
-  ImageF weights(xsize, ysize);
+  JXL_ASSIGN_OR_RETURN(ImageF weights, ImageF::Create(xsize, ysize));
   for (size_t y = 0; y < weights.ysize(); y++) {
     auto* row = weights.Row(y);
     for (size_t x = 0; x < weights.xsize(); x++) {
       row[x] = 1;
     }
   }
-  ImageF weights2(xsize2, ysize2);
+  JXL_ASSIGN_OR_RETURN(ImageF weights2, ImageF::Create(xsize2, ysize2));
   AntiUpsample(weights, &weights2);
 
   const size_t num_it = 3;
   for (size_t it = 0; it < num_it; ++it) {
     UpsampleImage(down, &up);
-    corr = LinComb<float>(1, orig, -1, up);
+    JXL_ASSIGN_OR_RETURN(corr, LinComb<float>(1, orig, -1, up));
     ElwiseMul(corr, weights, &corr);
     AntiUpsample(corr, &corr2);
     ElwiseDiv(corr2, weights2, &corr2);
 
-    down = LinComb<float>(1, down, 1, corr2);
+    JXL_ASSIGN_OR_RETURN(down, LinComb<float>(1, down, 1, corr2));
   }
 
   ReduceRinging(initial, mask, down);
@@ -670,124 +699,85 @@ void DownsampleImage2_Iterative(const ImageF& orig, ImageF* output) {
       output->Row(y)[x] = v;
     }
   }
+  return true;
 }
 
-void DownsampleImage2_Iterative(Image3F* opsin) {
+}  // namespace
+
+Status DownsampleImage2_Iterative(Image3F* opsin) {
   // Allocate extra space to avoid a reallocation when padding.
-  Image3F downsampled(DivCeil(opsin->xsize(), 2) + kBlockDim,
-                      DivCeil(opsin->ysize(), 2) + kBlockDim);
+  JXL_ASSIGN_OR_RETURN(Image3F downsampled,
+                       Image3F::Create(DivCeil(opsin->xsize(), 2) + kBlockDim,
+                                       DivCeil(opsin->ysize(), 2) + kBlockDim));
   downsampled.ShrinkTo(downsampled.xsize() - kBlockDim,
                        downsampled.ysize() - kBlockDim);
 
-  Image3F rgb(opsin->xsize(), opsin->ysize());
-  OpsinParams opsin_params;  // TODO: use the ones that are actually used
+  JXL_ASSIGN_OR_RETURN(Image3F rgb,
+                       Image3F::Create(opsin->xsize(), opsin->ysize()));
+  OpsinParams opsin_params;  // TODO(user): use the ones that are actually used
   opsin_params.Init(kDefaultIntensityTarget);
   OpsinToLinear(*opsin, Rect(rgb), nullptr, &rgb, opsin_params);
 
-  ImageF mask(opsin->xsize(), opsin->ysize());
+  JXL_ASSIGN_OR_RETURN(ImageF mask,
+                       ImageF::Create(opsin->xsize(), opsin->ysize()));
   ButteraugliParams butter_params;
-  ButteraugliComparator butter(rgb, butter_params);
-  butter.Mask(&mask);
-  ImageF mask_fuzzy(opsin->xsize(), opsin->ysize());
+  JXL_ASSIGN_OR_RETURN(std::unique_ptr<ButteraugliComparator> butter,
+                       ButteraugliComparator::Make(rgb, butter_params));
+  JXL_RETURN_IF_ERROR(butter->Mask(&mask));
+  JXL_ASSIGN_OR_RETURN(ImageF mask_fuzzy,
+                       ImageF::Create(opsin->xsize(), opsin->ysize()));
 
   for (size_t c = 0; c < 3; c++) {
-    DownsampleImage2_Iterative(opsin->Plane(c), &downsampled.Plane(c));
+    JXL_RETURN_IF_ERROR(
+        DownsampleImage2_Iterative(opsin->Plane(c), &downsampled.Plane(c)));
   }
   *opsin = std::move(downsampled);
+  return true;
 }
-}  // namespace
 
-Status DefaultEncoderHeuristics::LossyFrameHeuristics(
-    PassesEncoderState* enc_state, ModularFrameEncoder* modular_frame_encoder,
-    const ImageBundle* original_pixels, Image3F* opsin,
-    const JxlCmsInterface& cms, ThreadPool* pool, AuxOut* aux_out) {
-  PROFILER_ZONE("JxlLossyFrameHeuristics uninstrumented");
-
-  CompressParams& cparams = enc_state->cparams;
+Status LossyFrameHeuristics(const FrameHeader& frame_header,
+                            PassesEncoderState* enc_state,
+                            ModularFrameEncoder* modular_frame_encoder,
+                            const Image3F* original_pixels, Image3F* opsin,
+                            const Rect& rect, const JxlCmsInterface& cms,
+                            ThreadPool* pool, AuxOut* aux_out) {
+  const CompressParams& cparams = enc_state->cparams;
+  const bool streaming_mode = enc_state->streaming_mode;
+  const bool initialize_global_state = enc_state->initialize_global_state;
   PassesSharedState& shared = enc_state->shared;
-
-  // Compute parameters for noise synthesis.
-  if (shared.frame_header.flags & FrameHeader::kNoise) {
-    PROFILER_ZONE("enc GetNoiseParam");
-    if (cparams.photon_noise_iso == 0) {
-      // Don't start at zero amplitude since adding noise is expensive -- it
-      // significantly slows down decoding, and this is unlikely to
-      // completely go away even with advanced optimizations. After the
-      // kNoiseModelingRampUpDistanceRange we have reached the full level,
-      // i.e. noise is no longer represented by the compressed image, so we
-      // can add full noise by the noise modeling itself.
-      static const float kNoiseModelingRampUpDistanceRange = 0.6;
-      static const float kNoiseLevelAtStartOfRampUp = 0.25;
-      static const float kNoiseRampupStart = 1.0;
-      // TODO(user) test and properly select quality_coef with smooth
-      // filter
-      float quality_coef = 1.0f;
-      const float rampup = (cparams.butteraugli_distance - kNoiseRampupStart) /
-                           kNoiseModelingRampUpDistanceRange;
-      if (rampup < 1.0f) {
-        quality_coef = kNoiseLevelAtStartOfRampUp +
-                       (1.0f - kNoiseLevelAtStartOfRampUp) * rampup;
-      }
-      if (rampup < 0.0f) {
-        quality_coef = kNoiseRampupStart;
-      }
-      if (!GetNoiseParameter(*opsin, &shared.image_features.noise_params,
-                             quality_coef)) {
-        shared.frame_header.flags &= ~FrameHeader::kNoise;
-      }
-    }
-  }
-  if (enc_state->shared.frame_header.upsampling != 1 &&
-      !cparams.already_downsampled) {
-    // In VarDCT mode, LossyFrameHeuristics takes care of running downsampling
-    // after noise, if necessary.
-    if (cparams.resampling == 2) {
-      // TODO(lode): use the regular DownsampleImage, or adapt to the custom
-      // coefficients, if there is are custom upscaling coefficients in
-      // CustomTransformData
-      if (cparams.speed_tier <= SpeedTier::kSquirrel) {
-        // TODO(lode): DownsampleImage2_Iterative is currently too slow to
-        // be used for squirrel, make it faster, and / or enable it only for
-        // kitten.
-        DownsampleImage2_Iterative(opsin);
-      } else {
-        DownsampleImage2_Sharper(opsin);
-      }
-    } else {
-      DownsampleImage(opsin, cparams.resampling);
-    }
-    PadImageToBlockMultipleInPlace(opsin);
-  }
-
-  if (cparams.butteraugli_distance < 0) {
-    return JXL_FAILURE("Expected non-negative distance");
-  }
+  const FrameDimensions& frame_dim = shared.frame_dim;
+  ImageFeatures& image_features = shared.image_features;
+  DequantMatrices& matrices = shared.matrices;
+  Quantizer& quantizer = shared.quantizer;
+  ImageI& raw_quant_field = shared.raw_quant_field;
+  ColorCorrelationMap& cmap = shared.cmap;
+  AcStrategyImage& ac_strategy = shared.ac_strategy;
+  ImageB& epf_sharpness = shared.epf_sharpness;
+  BlockCtxMap& block_ctx_map = shared.block_ctx_map;
 
   // Find and subtract splines.
-  if (cparams.speed_tier <= SpeedTier::kSquirrel) {
-    // If we do already have them, they were passed upstream to EncodeFile.
-    if (!shared.image_features.splines.HasAny()) {
-      shared.image_features.splines = FindSplines(*opsin);
+  if (cparams.custom_splines.HasAny()) {
+    image_features.splines = cparams.custom_splines;
+  }
+  if (!streaming_mode && cparams.speed_tier <= SpeedTier::kSquirrel) {
+    if (!cparams.custom_splines.HasAny()) {
+      image_features.splines = FindSplines(*opsin);
     }
-    JXL_RETURN_IF_ERROR(shared.image_features.splines.InitializeDrawCache(
-        opsin->xsize(), opsin->ysize(), shared.cmap));
-    shared.image_features.splines.SubtractFrom(opsin);
+    JXL_RETURN_IF_ERROR(image_features.splines.InitializeDrawCache(
+        opsin->xsize(), opsin->ysize(), cmap));
+    image_features.splines.SubtractFrom(opsin);
   }
 
   // Find and subtract patches/dots.
-  if (ApplyOverride(cparams.patches,
+  if (!streaming_mode &&
+      ApplyOverride(cparams.patches,
                     cparams.speed_tier <= SpeedTier::kSquirrel)) {
-    FindBestPatchDictionary(*opsin, enc_state, cms, pool, aux_out);
-    PatchDictionaryEncoder::SubtractFrom(shared.image_features.patches, opsin);
+    JXL_RETURN_IF_ERROR(
+        FindBestPatchDictionary(*opsin, enc_state, cms, pool, aux_out));
+    PatchDictionaryEncoder::SubtractFrom(image_features.patches, opsin);
   }
 
-  static const float kAcQuant = 0.79f;
   const float quant_dc = InitialQuantDC(cparams.butteraugli_distance);
-  Quantizer& quantizer = enc_state->shared.quantizer;
-  // We don't know the quant field yet, but for computing the global scale
-  // assuming that it will be the same as for Falcon mode is good enough.
-  quantizer.ComputeGlobalScaleAndQuant(
-      quant_dc, kAcQuant / cparams.butteraugli_distance, 0);
 
   // TODO(veluca): we can now run all the code from here to FindBestQuantizer
   // (excluded) one rect at a time. Do that.
@@ -809,128 +799,140 @@ Status DefaultEncoderHeuristics::LossyFrameHeuristics(
   // output: Gaborished XYB, CfL, ACS, raw quant field, EPF control field.
 
   ArControlFieldHeuristics ar_heuristics;
-  AcStrategyHeuristics acs_heuristics;
+  AcStrategyHeuristics acs_heuristics(cparams);
   CfLHeuristics cfl_heuristics;
-
-  if (!opsin->xsize()) {
-    JXL_ASSERT(HandlesColorConversion(cparams, *original_pixels));
-    *opsin = Image3F(RoundUpToBlockDim(original_pixels->xsize()),
-                     RoundUpToBlockDim(original_pixels->ysize()));
-    opsin->ShrinkTo(original_pixels->xsize(), original_pixels->ysize());
-    ToXYB(*original_pixels, pool, opsin, cms, /*linear=*/nullptr);
-    PadImageToBlockMultipleInPlace(opsin);
-  }
+  ImageF initial_quant_field;
+  ImageF initial_quant_masking;
+  ImageF initial_quant_masking1x1;
 
   // Compute an initial estimate of the quantization field.
   // Call InitialQuantField only in Hare mode or slower. Otherwise, rely
   // on simple heuristics in FindBestAcStrategy, or set a constant for Falcon
   // mode.
-  if (cparams.speed_tier > SpeedTier::kHare || cparams.uniform_quant > 0) {
-    enc_state->initial_quant_field =
-        ImageF(shared.frame_dim.xsize_blocks, shared.frame_dim.ysize_blocks);
-    enc_state->initial_quant_masking =
-        ImageF(shared.frame_dim.xsize_blocks, shared.frame_dim.ysize_blocks);
-    float q = cparams.uniform_quant > 0
-                  ? cparams.uniform_quant
-                  : kAcQuant / cparams.butteraugli_distance;
-    FillImage(q, &enc_state->initial_quant_field);
-    FillImage(1.0f / (q + 0.001f), &enc_state->initial_quant_masking);
+  if (cparams.speed_tier > SpeedTier::kHare) {
+    JXL_ASSIGN_OR_RETURN(
+        initial_quant_field,
+        ImageF::Create(frame_dim.xsize_blocks, frame_dim.ysize_blocks));
+    JXL_ASSIGN_OR_RETURN(
+        initial_quant_masking,
+        ImageF::Create(frame_dim.xsize_blocks, frame_dim.ysize_blocks));
+    float q = 0.79 / cparams.butteraugli_distance;
+    FillImage(q, &initial_quant_field);
+    FillImage(1.0f / (q + 0.001f), &initial_quant_masking);
+    quantizer.ComputeGlobalScaleAndQuant(quant_dc, q, 0);
   } else {
     // Call this here, as it relies on pre-gaborish values.
     float butteraugli_distance_for_iqf = cparams.butteraugli_distance;
-    if (!shared.frame_header.loop_filter.gab) {
+    if (!frame_header.loop_filter.gab) {
       butteraugli_distance_for_iqf *= 0.73f;
     }
-    enc_state->initial_quant_field = InitialQuantField(
-        butteraugli_distance_for_iqf, *opsin, shared.frame_dim, pool, 1.0f,
-        &enc_state->initial_quant_masking);
-    quantizer.SetQuantField(quant_dc, enc_state->initial_quant_field, nullptr);
+    JXL_ASSIGN_OR_RETURN(
+        initial_quant_field,
+        InitialQuantField(butteraugli_distance_for_iqf, *opsin, rect, pool,
+                          1.0f, &initial_quant_masking,
+                          &initial_quant_masking1x1));
+    float q = 0.39 / cparams.butteraugli_distance;
+    quantizer.ComputeGlobalScaleAndQuant(quant_dc, q, 0);
   }
 
   // TODO(veluca): do something about animations.
 
   // Apply inverse-gaborish.
-  if (shared.frame_header.loop_filter.gab) {
-    GaborishInverse(opsin, 0.9908511000000001f, pool);
+  if (frame_header.loop_filter.gab) {
+    // Unsure why better to do some more gaborish on X and B than Y.
+    float weight[3] = {
+        1.0036278514398933f,
+        0.99406123118127299f,
+        0.99719338015886894f,
+    };
+    JXL_RETURN_IF_ERROR(GaborishInverse(opsin, rect, weight, pool));
   }
 
-  FindBestDequantMatrices(cparams, *opsin, modular_frame_encoder,
-                          &enc_state->shared.matrices);
+  if (initialize_global_state) {
+    JXL_RETURN_IF_ERROR(
+        FindBestDequantMatrices(cparams, modular_frame_encoder, &matrices));
+  }
 
-  cfl_heuristics.Init(*opsin);
-  acs_heuristics.Init(*opsin, enc_state);
+  JXL_RETURN_IF_ERROR(cfl_heuristics.Init(rect));
+  acs_heuristics.Init(*opsin, rect, initial_quant_field, initial_quant_masking,
+                      initial_quant_masking1x1, &matrices);
 
+  std::atomic<bool> has_error{false};
   auto process_tile = [&](const uint32_t tid, const size_t thread) {
-    size_t n_enc_tiles =
-        DivCeil(enc_state->shared.frame_dim.xsize_blocks, kEncTileDimInBlocks);
+    if (has_error) return;
+    size_t n_enc_tiles = DivCeil(frame_dim.xsize_blocks, kEncTileDimInBlocks);
     size_t tx = tid % n_enc_tiles;
     size_t ty = tid / n_enc_tiles;
     size_t by0 = ty * kEncTileDimInBlocks;
-    size_t by1 = std::min((ty + 1) * kEncTileDimInBlocks,
-                          enc_state->shared.frame_dim.ysize_blocks);
+    size_t by1 =
+        std::min((ty + 1) * kEncTileDimInBlocks, frame_dim.ysize_blocks);
     size_t bx0 = tx * kEncTileDimInBlocks;
-    size_t bx1 = std::min((tx + 1) * kEncTileDimInBlocks,
-                          enc_state->shared.frame_dim.xsize_blocks);
+    size_t bx1 =
+        std::min((tx + 1) * kEncTileDimInBlocks, frame_dim.xsize_blocks);
     Rect r(bx0, by0, bx1 - bx0, by1 - by0);
 
     // For speeds up to Wombat, we only compute the color correlation map
     // once we know the transform type and the quantization map.
     if (cparams.speed_tier <= SpeedTier::kSquirrel) {
-      cfl_heuristics.ComputeTile(r, *opsin, enc_state->shared.matrices,
+      cfl_heuristics.ComputeTile(r, *opsin, rect, matrices,
                                  /*ac_strategy=*/nullptr,
+                                 /*raw_quant_field=*/nullptr,
                                  /*quantizer=*/nullptr, /*fast=*/false, thread,
-                                 &enc_state->shared.cmap);
+                                 &cmap);
     }
 
     // Choose block sizes.
-    acs_heuristics.ProcessRect(r);
+    acs_heuristics.ProcessRect(r, cmap, &ac_strategy);
 
     // Choose amount of post-processing smoothing.
     // TODO(veluca): should this go *after* AdjustQuantField?
-    ar_heuristics.RunRect(r, *opsin, enc_state, thread);
+    if (!ar_heuristics.RunRect(cparams, frame_header, r, *opsin, rect,
+                               initial_quant_field, ac_strategy, &epf_sharpness,
+                               thread)) {
+      has_error = true;
+      return;
+    }
 
     // Always set the initial quant field, so we can compute the CfL map with
     // more accuracy. The initial quant field might change in slower modes, but
     // adjusting the quant field with butteraugli when all the other encoding
     // parameters are fixed is likely a more reliable choice anyway.
-    AdjustQuantField(enc_state->shared.ac_strategy, r,
-                     &enc_state->initial_quant_field);
-    quantizer.SetQuantFieldRect(enc_state->initial_quant_field, r,
-                                &enc_state->shared.raw_quant_field);
+    AdjustQuantField(ac_strategy, r, cparams.butteraugli_distance,
+                     &initial_quant_field);
+    quantizer.SetQuantFieldRect(initial_quant_field, r, &raw_quant_field);
 
     // Compute a non-default CfL map if we are at Hare speed, or slower.
     if (cparams.speed_tier <= SpeedTier::kHare) {
       cfl_heuristics.ComputeTile(
-          r, *opsin, enc_state->shared.matrices, &enc_state->shared.ac_strategy,
-          &enc_state->shared.quantizer,
-          /*fast=*/cparams.speed_tier >= SpeedTier::kWombat, thread,
-          &enc_state->shared.cmap);
+          r, *opsin, rect, matrices, &ac_strategy, &raw_quant_field, &quantizer,
+          /*fast=*/cparams.speed_tier >= SpeedTier::kWombat, thread, &cmap);
     }
   };
   JXL_RETURN_IF_ERROR(RunOnPool(
       pool, 0,
-      DivCeil(enc_state->shared.frame_dim.xsize_blocks, kEncTileDimInBlocks) *
-          DivCeil(enc_state->shared.frame_dim.ysize_blocks,
-                  kEncTileDimInBlocks),
+      DivCeil(frame_dim.xsize_blocks, kEncTileDimInBlocks) *
+          DivCeil(frame_dim.ysize_blocks, kEncTileDimInBlocks),
       [&](const size_t num_threads) {
         ar_heuristics.PrepareForThreads(num_threads);
         cfl_heuristics.PrepareForThreads(num_threads);
         return true;
       },
       process_tile, "Enc Heuristics"));
+  if (has_error) return JXL_FAILURE("Enc Heuristics failed");
 
-  acs_heuristics.Finalize(aux_out);
-  if (cparams.speed_tier <= SpeedTier::kHare) {
-    cfl_heuristics.ComputeDC(/*fast=*/cparams.speed_tier >= SpeedTier::kWombat,
-                             &enc_state->shared.cmap);
-  }
+  JXL_RETURN_IF_ERROR(acs_heuristics.Finalize(frame_dim, ac_strategy, aux_out));
 
   // Refine quantization levels.
-  FindBestQuantizer(original_pixels, *opsin, enc_state, cms, pool, aux_out);
+  if (!streaming_mode) {
+    JXL_RETURN_IF_ERROR(FindBestQuantizer(frame_header, original_pixels, *opsin,
+                                          initial_quant_field, enc_state, cms,
+                                          pool, aux_out));
+  }
 
   // Choose a context model that depends on the amount of quantization for AC.
-  if (cparams.speed_tier < SpeedTier::kFalcon) {
-    FindBestBlockEntropyModel(*enc_state);
+  if (cparams.speed_tier < SpeedTier::kFalcon && initialize_global_state) {
+    FindBestBlockEntropyModel(cparams, raw_quant_field, ac_strategy,
+                              &block_ctx_map);
   }
   return true;
 }
