@@ -7,22 +7,15 @@
 
 #include <stdlib.h>
 
-#include <algorithm>
 #include <cmath>
-#include <limits>
-#include <utility>
 
-#include "lib/jxl/base/bits.h"
 #include "lib/jxl/base/common.h"
 #include "lib/jxl/base/status.h"
-#include "lib/jxl/dct_scales.h"
 #include "lib/jxl/enc_aux_out.h"
 #include "lib/jxl/enc_bit_writer.h"
 #include "lib/jxl/enc_modular.h"
 #include "lib/jxl/fields.h"
-#include "lib/jxl/image.h"
 #include "lib/jxl/modular/encoding/encoding.h"
-#include "lib/jxl/modular/options.h"
 
 namespace jxl {
 
@@ -95,8 +88,8 @@ Status EncodeQuant(const QuantEncoding& encoding, size_t idx, size_t size_x,
       break;
     }
     case QuantEncoding::kQuantModeRAW: {
-      ModularFrameEncoder::EncodeQuantTable(size_x, size_y, writer, encoding,
-                                            idx, modular_frame_encoder);
+      JXL_RETURN_IF_ERROR(ModularFrameEncoder::EncodeQuantTable(
+          size_x, size_y, writer, encoding, idx, modular_frame_encoder));
       break;
     }
     case QuantEncoding::kQuantModeAFV: {
@@ -116,11 +109,11 @@ Status EncodeQuant(const QuantEncoding& encoding, size_t idx, size_t size_x,
 
 }  // namespace
 
-Status DequantMatricesEncode(const DequantMatrices* matrices, BitWriter* writer,
+Status DequantMatricesEncode(const DequantMatrices& matrices, BitWriter* writer,
                              size_t layer, AuxOut* aux_out,
                              ModularFrameEncoder* modular_frame_encoder) {
   bool all_default = true;
-  const std::vector<QuantEncoding>& encodings = matrices->encodings();
+  const std::vector<QuantEncoding>& encodings = matrices.encodings();
 
   for (size_t i = 0; i < encodings.size(); i++) {
     if (encodings[i].mode != QuantEncoding::kQuantModeLibrary ||
@@ -142,11 +135,11 @@ Status DequantMatricesEncode(const DequantMatrices* matrices, BitWriter* writer,
   return true;
 }
 
-Status DequantMatricesEncodeDC(const DequantMatrices* matrices,
+Status DequantMatricesEncodeDC(const DequantMatrices& matrices,
                                BitWriter* writer, size_t layer,
                                AuxOut* aux_out) {
   bool all_default = true;
-  const float* dc_quant = matrices->DCQuants();
+  const float* dc_quant = matrices.DCQuants();
   for (size_t c = 0; c < 3; c++) {
     if (dc_quant[c] != kDCQuant[c]) {
       all_default = false;
@@ -167,7 +160,7 @@ void DequantMatricesSetCustomDC(DequantMatrices* matrices, const float* dc) {
   matrices->SetDCQuant(dc);
   // Roundtrip encode/decode DC to ensure same values as decoder.
   BitWriter writer;
-  JXL_CHECK(DequantMatricesEncodeDC(matrices, &writer, 0, nullptr));
+  JXL_CHECK(DequantMatricesEncodeDC(*matrices, &writer, 0, nullptr));
   writer.ZeroPadToByte();
   BitReader br(writer.GetSpan());
   // Called only in the encoder: should fail only for programmer errors.
@@ -187,7 +180,7 @@ void DequantMatricesRoundtrip(DequantMatrices* matrices) {
   // Do not pass modular en/decoder, as they only change entropy and not
   // values.
   BitWriter writer;
-  JXL_CHECK(DequantMatricesEncode(matrices, &writer, 0, nullptr));
+  JXL_CHECK(DequantMatricesEncode(*matrices, &writer, 0, nullptr));
   writer.ZeroPadToByte();
   BitReader br(writer.GetSpan());
   // Called only in the encoder: should fail only for programmer errors.
@@ -195,19 +188,20 @@ void DequantMatricesRoundtrip(DequantMatrices* matrices) {
   JXL_CHECK(br.Close());
 }
 
-void DequantMatricesSetCustom(DequantMatrices* matrices,
-                              const std::vector<QuantEncoding>& encodings,
-                              ModularFrameEncoder* encoder) {
+Status DequantMatricesSetCustom(DequantMatrices* matrices,
+                                const std::vector<QuantEncoding>& encodings,
+                                ModularFrameEncoder* encoder) {
   JXL_ASSERT(encodings.size() == DequantMatrices::kNum);
   matrices->SetEncodings(encodings);
   for (size_t i = 0; i < encodings.size(); i++) {
     if (encodings[i].mode == QuantEncodingInternal::kQuantModeRAW) {
-      encoder->AddQuantTable(DequantMatrices::required_size_x[i] * kBlockDim,
-                             DequantMatrices::required_size_y[i] * kBlockDim,
-                             encodings[i], i);
+      JXL_RETURN_IF_ERROR(encoder->AddQuantTable(
+          DequantMatrices::required_size_x[i] * kBlockDim,
+          DequantMatrices::required_size_y[i] * kBlockDim, encodings[i], i));
     }
   }
   DequantMatricesRoundtrip(matrices);
+  return true;
 }
 
 }  // namespace jxl
