@@ -8,12 +8,18 @@
 #include <stdlib.h>
 #include <string.h> /* for memset, memcpy */
 
+#include <algorithm>
+#include <cstddef>
+#include <cstdint>
 #include <deque>
 #include <string>
 #include <vector>
 
 #include "lib/jxl/base/bits.h"
-#include "lib/jxl/common.h"
+#include "lib/jxl/base/byte_order.h"
+#include "lib/jxl/base/common.h"
+#include "lib/jxl/frame_dimensions.h"
+#include "lib/jxl/image_bundle.h"
 #include "lib/jxl/jpeg/dec_jpeg_serialization_state.h"
 #include "lib/jxl/jpeg/jpeg_data.h"
 
@@ -34,12 +40,9 @@ const int kJpegPrecision = 8;
 // JpegBitWriter: buffer size
 const size_t kJpegBitWriterChunkSize = 16384;
 
-// DCTCodingState: maximum number of correction bits to buffer
-const int kJPEGMaxCorrectionBits = 1u << 16;
-
 // Returns non-zero if and only if x has a zero byte, i.e. one of
 // x & 0xff, x & 0xff00, ..., x & 0xff00000000000000 is zero.
-static JXL_INLINE uint64_t HasZeroByte(uint64_t x) {
+JXL_INLINE uint64_t HasZeroByte(uint64_t x) {
   return (x - 0x0101010101010101ULL) & ~x & 0x8080808080808080ULL;
 }
 
@@ -54,7 +57,7 @@ void JpegBitWriterInit(JpegBitWriter* bw,
   bw->data = bw->chunk.buffer->data();
 }
 
-static JXL_NOINLINE void SwapBuffer(JpegBitWriter* bw) {
+JXL_NOINLINE void SwapBuffer(JpegBitWriter* bw) {
   bw->chunk.len = bw->pos;
   bw->output->emplace_back(std::move(bw->chunk));
   bw->chunk = OutputChunk(kJpegBitWriterChunkSize);
@@ -62,7 +65,7 @@ static JXL_NOINLINE void SwapBuffer(JpegBitWriter* bw) {
   bw->pos = 0;
 }
 
-static JXL_INLINE void Reserve(JpegBitWriter* bw, size_t n_bytes) {
+JXL_INLINE void Reserve(JpegBitWriter* bw, size_t n_bytes) {
   if (JXL_UNLIKELY((bw->pos + n_bytes) > kJpegBitWriterChunkSize)) {
     SwapBuffer(bw);
   }
@@ -74,19 +77,20 @@ static JXL_INLINE void Reserve(JpegBitWriter* bw, size_t n_bytes) {
  * This method is "careless" - caller must make sure that there is enough
  * space in the output buffer. Emits up to 2 bytes to buffer.
  */
-static JXL_INLINE void EmitByte(JpegBitWriter* bw, int byte) {
-  bw->data[bw->pos++] = byte;
-  if (byte == 0xFF) bw->data[bw->pos++] = 0;
+JXL_INLINE void EmitByte(JpegBitWriter* bw, int byte) {
+  bw->data[bw->pos] = byte;
+  bw->data[bw->pos + 1] = 0;
+  bw->pos += (byte != 0xFF ? 1 : 2);
 }
 
-static JXL_INLINE void DischargeBitBuffer(JpegBitWriter* bw) {
-  // At this point we are ready to emit the most significant 6 bytes of
-  // put_buffer_ to the output.
+JXL_INLINE void DischargeBitBuffer(JpegBitWriter* bw, int nbits,
+                                   uint64_t bits) {
+  // At this point we are ready to emit the put_buffer to the output.
   // The JPEG format requires that after every 0xff byte in the entropy
   // coded section, there is a zero byte, therefore we first check if any of
-  // the 6 most significant bytes of put_buffer_ is 0xFF.
-  Reserve(bw, 12);
-  if (HasZeroByte(~bw->put_buffer | 0xFFFF)) {
+  // the 8 bytes of put_buffer is 0xFF.
+  bw->put_buffer |= (bits >> -bw->put_bits);
+  if (JXL_UNLIKELY(HasZeroByte(~bw->put_buffer))) {
     // We have a 0xFF byte somewhere, examine each byte and append a zero
     // byte if necessary.
     EmitByte(bw, (bw->put_buffer >> 56) & 0xFF);
@@ -95,32 +99,31 @@ static JXL_INLINE void DischargeBitBuffer(JpegBitWriter* bw) {
     EmitByte(bw, (bw->put_buffer >> 32) & 0xFF);
     EmitByte(bw, (bw->put_buffer >> 24) & 0xFF);
     EmitByte(bw, (bw->put_buffer >> 16) & 0xFF);
+    EmitByte(bw, (bw->put_buffer >> 8) & 0xFF);
+    EmitByte(bw, (bw->put_buffer) & 0xFF);
   } else {
-    // We don't have any 0xFF bytes, output all 6 bytes without checking.
-    bw->data[bw->pos] = (bw->put_buffer >> 56) & 0xFF;
-    bw->data[bw->pos + 1] = (bw->put_buffer >> 48) & 0xFF;
-    bw->data[bw->pos + 2] = (bw->put_buffer >> 40) & 0xFF;
-    bw->data[bw->pos + 3] = (bw->put_buffer >> 32) & 0xFF;
-    bw->data[bw->pos + 4] = (bw->put_buffer >> 24) & 0xFF;
-    bw->data[bw->pos + 5] = (bw->put_buffer >> 16) & 0xFF;
-    bw->pos += 6;
+    // We don't have any 0xFF bytes, output all 8 bytes without checking.
+    StoreBE64(bw->put_buffer, bw->data + bw->pos);
+    bw->pos += 8;
   }
-  bw->put_buffer <<= 48;
-  bw->put_bits += 48;
+
+  bw->put_bits += 64;
+  bw->put_buffer = bits << bw->put_bits;
 }
 
-static JXL_INLINE void WriteBits(JpegBitWriter* bw, int nbits, uint64_t bits) {
-  // This is an optimization; if everything goes well,
-  // then |nbits| is positive; if non-existing Huffman symbol is going to be
-  // encoded, its length should be zero; later encoder could check the
-  // "health" of JpegBitWriter.
-  if (nbits == 0) {
-    bw->healthy = false;
-    return;
-  }
+JXL_INLINE void WriteBits(JpegBitWriter* bw, int nbits, uint64_t bits) {
+  JXL_DASSERT(nbits > 0);
   bw->put_bits -= nbits;
-  bw->put_buffer |= (bits << bw->put_bits);
-  if (bw->put_bits <= 16) DischargeBitBuffer(bw);
+  if (JXL_UNLIKELY(bw->put_bits < 0)) {
+    if (JXL_UNLIKELY(nbits > 64)) {
+      bw->put_bits += nbits;
+      bw->healthy = false;
+    } else {
+      DischargeBitBuffer(bw, nbits, bits);
+    }
+  } else {
+    bw->put_buffer |= (bits << bw->put_bits);
+  }
 }
 
 void EmitMarker(JpegBitWriter* bw, int marker) {
@@ -181,61 +184,89 @@ void DCTCodingStateInit(DCTCodingState* s) {
   s->eob_run_ = 0;
   s->cur_ac_huff_ = nullptr;
   s->refinement_bits_.clear();
-  s->refinement_bits_.reserve(kJPEGMaxCorrectionBits);
+  s->refinement_bits_.reserve(64);
 }
 
-enum OutputModes {
-  kModeHistogram,
-  kModeWrite,
-};
+JXL_INLINE void WriteSymbol(int symbol, HuffmanCodeTable* table,
+                            JpegBitWriter* bw) {
+  WriteBits(bw, table->depth[symbol], table->code[symbol]);
+}
 
-template <int kOutputMode>
-static JXL_INLINE void WriteSymbol(int symbol, HuffmanCodeTable* table,
-                                   JpegBitWriter* bw) {
-  if (kOutputMode == OutputModes::kModeHistogram) {
-    ++table->depth[symbol];
-  } else {
-    WriteBits(bw, table->depth[symbol], table->code[symbol]);
-  }
+JXL_INLINE void WriteSymbolBits(int symbol, HuffmanCodeTable* table,
+                                JpegBitWriter* bw, int nbits, uint64_t bits) {
+  WriteBits(bw, nbits + table->depth[symbol],
+            bits | (table->code[symbol] << nbits));
 }
 
 // Emit all buffered data to the bit stream using the given Huffman code and
 // bit writer.
-template <int kOutputMode>
-static JXL_INLINE void Flush(DCTCodingState* s, JpegBitWriter* bw) {
+JXL_INLINE void Flush(DCTCodingState* s, JpegBitWriter* bw) {
   if (s->eob_run_ > 0) {
+    Reserve(bw, 16);
     int nbits = FloorLog2Nonzero<uint32_t>(s->eob_run_);
     int symbol = nbits << 4u;
-    WriteSymbol<kOutputMode>(symbol, s->cur_ac_huff_, bw);
+    WriteSymbol(symbol, s->cur_ac_huff_, bw);
     if (nbits > 0) {
       WriteBits(bw, nbits, s->eob_run_ & ((1 << nbits) - 1));
     }
     s->eob_run_ = 0;
   }
-  for (size_t i = 0; i < s->refinement_bits_.size(); ++i) {
-    WriteBits(bw, 1, s->refinement_bits_[i]);
+  const size_t kStride = 124;  // (515 - 16) / 2 / 2
+  size_t num_words = s->refinement_bits_count_ >> 4;
+  size_t i = 0;
+  while (i < num_words) {
+    size_t limit = std::min(i + kStride, num_words);
+    Reserve(bw, 512);
+    for (; i < limit; ++i) {
+      WriteBits(bw, 16, s->refinement_bits_[i]);
+    }
+  }
+  Reserve(bw, 16);
+  size_t tail = s->refinement_bits_count_ & 0xF;
+  if (tail) {
+    WriteBits(bw, tail, s->refinement_bits_.back());
   }
   s->refinement_bits_.clear();
+  s->refinement_bits_count_ = 0;
 }
 
 // Buffer some more data at the end-of-band (the last non-zero or newly
 // non-zero coefficient within the [Ss, Se] spectral band).
-template <int kOutputMode>
-static JXL_INLINE void BufferEndOfBand(DCTCodingState* s,
-                                       HuffmanCodeTable* ac_huff,
-                                       const std::vector<int>* new_bits,
-                                       JpegBitWriter* bw) {
+JXL_INLINE void BufferEndOfBand(DCTCodingState* s, HuffmanCodeTable* ac_huff,
+                                const int* new_bits_array,
+                                size_t new_bits_count, JpegBitWriter* bw) {
   if (s->eob_run_ == 0) {
     s->cur_ac_huff_ = ac_huff;
   }
   ++s->eob_run_;
-  if (new_bits) {
-    s->refinement_bits_.insert(s->refinement_bits_.end(), new_bits->begin(),
-                               new_bits->end());
+  if (new_bits_count) {
+    uint64_t new_bits = 0;
+    for (size_t i = 0; i < new_bits_count; ++i) {
+      new_bits = (new_bits << 1) | new_bits_array[i];
+    }
+    size_t tail = s->refinement_bits_count_ & 0xF;
+    if (tail) {  // First stuff the tail item
+      size_t stuff_bits_count = std::min(16 - tail, new_bits_count);
+      uint16_t stuff_bits = new_bits >> (new_bits_count - stuff_bits_count);
+      stuff_bits &= ((1u << stuff_bits_count) - 1);
+      s->refinement_bits_.back() =
+          (s->refinement_bits_.back() << stuff_bits_count) | stuff_bits;
+      new_bits_count -= stuff_bits_count;
+      s->refinement_bits_count_ += stuff_bits_count;
+    }
+    while (new_bits_count >= 16) {
+      s->refinement_bits_.push_back(new_bits >> (new_bits_count - 16));
+      new_bits_count -= 16;
+      s->refinement_bits_count_ += 16;
+    }
+    if (new_bits_count) {
+      s->refinement_bits_.push_back(new_bits & ((1u << new_bits_count) - 1));
+      s->refinement_bits_count_ += new_bits_count;
+    }
   }
-  if (s->eob_run_ == 0x7FFF ||
-      s->refinement_bits_.size() > kJPEGMaxCorrectionBits - kDCTBlockSize + 1) {
-    Flush<kOutputMode>(s, bw);
+
+  if (s->eob_run_ == 0x7FFF) {
+    Flush(s, bw);
   }
 }
 
@@ -378,10 +409,11 @@ bool EncodeDHT(const JPEGData& jpg, SerializationState* state) {
       huff_table = &state->dc_huff_table[index];
     }
     // TODO(eustas): cache
-    // TODO(eustas): set up non-existing symbols
+    huff_table->InitDepths(127);
     if (!BuildHuffmanCodeTable(huff, huff_table)) {
       return false;
     }
+    huff_table->initialized = true;
     size_t total_count = 0;
     size_t max_length = 0;
     for (size_t i = 0; i < huff.counts.size(); ++i) {
@@ -480,62 +512,72 @@ bool EncodeInterMarkerData(const JPEGData& jpg, SerializationState* state) {
   return true;
 }
 
-template <int kOutputMode>
 bool EncodeDCTBlockSequential(const coeff_t* coeffs, HuffmanCodeTable* dc_huff,
                               HuffmanCodeTable* ac_huff, int num_zero_runs,
                               coeff_t* last_dc_coeff, JpegBitWriter* bw) {
   coeff_t temp2;
   coeff_t temp;
+  coeff_t litmus = 0;
   temp2 = coeffs[0];
   temp = temp2 - *last_dc_coeff;
   *last_dc_coeff = temp2;
-  temp2 = temp;
-  if (temp < 0) {
-    temp = -temp;
-    if (temp < 0) return false;
-    temp2--;
-  }
-  int dc_nbits = (temp == 0) ? 0 : (FloorLog2Nonzero<uint32_t>(temp) + 1);
-  WriteSymbol<kOutputMode>(dc_nbits, dc_huff, bw);
+  temp2 = temp >> (8 * sizeof(coeff_t) - 1);
+  temp += temp2;
+  temp2 ^= temp;
+
+  int dc_nbits = (temp2 == 0) ? 0 : (FloorLog2Nonzero<uint32_t>(temp2) + 1);
+  WriteSymbol(dc_nbits, dc_huff, bw);
+#if false
+  // If the input is corrupt, this could be triggered. Checking is
+  // costly though, so it makes more sense to avoid this branch.
+  // (producing a corrupt JPEG when the input is corrupt, instead
+  // of catching it and returning error)
   if (dc_nbits >= 12) return false;
-  if (dc_nbits > 0) {
-    WriteBits(bw, dc_nbits, temp2 & ((1u << dc_nbits) - 1));
+#endif
+  if (dc_nbits) {
+    WriteBits(bw, dc_nbits, temp & ((1u << dc_nbits) - 1));
   }
-  int r = 0;
-  for (int k = 1; k < 64; ++k) {
-    if ((temp = coeffs[kJPEGNaturalOrder[k]]) == 0) {
+  int16_t r = 0;
+
+  for (size_t i = 1; i < 64; i++) {
+    if ((temp = coeffs[kJPEGNaturalOrder[i]]) == 0) {
       r++;
-      continue;
-    }
-    if (temp < 0) {
-      temp = -temp;
-      if (temp < 0) return false;
-      temp2 = ~temp;
     } else {
-      temp2 = temp;
+      temp2 = temp >> (8 * sizeof(coeff_t) - 1);
+      temp += temp2;
+      temp2 ^= temp;
+      if (JXL_UNLIKELY(r > 15)) {
+        WriteSymbol(0xf0, ac_huff, bw);
+        r -= 16;
+        if (r > 15) {
+          WriteSymbol(0xf0, ac_huff, bw);
+          r -= 16;
+        }
+        if (r > 15) {
+          WriteSymbol(0xf0, ac_huff, bw);
+          r -= 16;
+        }
+      }
+      litmus |= temp2;
+      int ac_nbits =
+          FloorLog2Nonzero<uint32_t>(static_cast<uint16_t>(temp2)) + 1;
+      int symbol = (r << 4u) + ac_nbits;
+      WriteSymbolBits(symbol, ac_huff, bw, ac_nbits,
+                      temp & ((1 << ac_nbits) - 1));
+      r = 0;
     }
-    while (r > 15) {
-      WriteSymbol<kOutputMode>(0xf0, ac_huff, bw);
-      r -= 16;
-    }
-    int ac_nbits = FloorLog2Nonzero<uint32_t>(temp) + 1;
-    if (ac_nbits >= 16) return false;
-    int symbol = (r << 4u) + ac_nbits;
-    WriteSymbol<kOutputMode>(symbol, ac_huff, bw);
-    WriteBits(bw, ac_nbits, temp2 & ((1 << ac_nbits) - 1));
-    r = 0;
   }
+
   for (int i = 0; i < num_zero_runs; ++i) {
-    WriteSymbol<kOutputMode>(0xf0, ac_huff, bw);
+    WriteSymbol(0xf0, ac_huff, bw);
     r -= 16;
   }
   if (r > 0) {
-    WriteSymbol<kOutputMode>(0, ac_huff, bw);
+    WriteSymbol(0, ac_huff, bw);
   }
-  return true;
+  return (litmus >= 0);
 }
 
-template <int kOutputMode>
 bool EncodeDCTBlockProgressive(const coeff_t* coeffs, HuffmanCodeTable* dc_huff,
                                HuffmanCodeTable* ac_huff, int Ss, int Se,
                                int Al, int num_zero_runs,
@@ -555,8 +597,8 @@ bool EncodeDCTBlockProgressive(const coeff_t* coeffs, HuffmanCodeTable* dc_huff,
       temp2--;
     }
     int nbits = (temp == 0) ? 0 : (FloorLog2Nonzero<uint32_t>(temp) + 1);
-    WriteSymbol<kOutputMode>(nbits, dc_huff, bw);
-    if (nbits > 0) {
+    WriteSymbol(nbits, dc_huff, bw);
+    if (nbits) {
       WriteBits(bw, nbits, temp2 & ((1 << nbits) - 1));
     }
     ++Ss;
@@ -583,34 +625,33 @@ bool EncodeDCTBlockProgressive(const coeff_t* coeffs, HuffmanCodeTable* dc_huff,
       r++;
       continue;
     }
-    Flush<kOutputMode>(coding_state, bw);
+    Flush(coding_state, bw);
     while (r > 15) {
-      WriteSymbol<kOutputMode>(0xf0, ac_huff, bw);
+      WriteSymbol(0xf0, ac_huff, bw);
       r -= 16;
     }
     int nbits = FloorLog2Nonzero<uint32_t>(temp) + 1;
     int symbol = (r << 4u) + nbits;
-    WriteSymbol<kOutputMode>(symbol, ac_huff, bw);
+    WriteSymbol(symbol, ac_huff, bw);
     WriteBits(bw, nbits, temp2 & ((1 << nbits) - 1));
     r = 0;
   }
   if (num_zero_runs > 0) {
-    Flush<kOutputMode>(coding_state, bw);
+    Flush(coding_state, bw);
     for (int i = 0; i < num_zero_runs; ++i) {
-      WriteSymbol<kOutputMode>(0xf0, ac_huff, bw);
+      WriteSymbol(0xf0, ac_huff, bw);
       r -= 16;
     }
   }
   if (r > 0) {
-    BufferEndOfBand<kOutputMode>(coding_state, ac_huff, nullptr, bw);
+    BufferEndOfBand(coding_state, ac_huff, nullptr, 0, bw);
     if (!eob_run_allowed) {
-      Flush<kOutputMode>(coding_state, bw);
+      Flush(coding_state, bw);
     }
   }
   return true;
 }
 
-template <int kOutputMode>
 bool EncodeRefinementBits(const coeff_t* coeffs, HuffmanCodeTable* ac_huff,
                           int Ss, int Se, int Al, DCTCodingState* coding_state,
                           JpegBitWriter* bw) {
@@ -633,64 +674,48 @@ bool EncodeRefinementBits(const coeff_t* coeffs, HuffmanCodeTable* ac_huff,
     }
   }
   int r = 0;
-  std::vector<int> refinement_bits;
-  refinement_bits.reserve(kDCTBlockSize);
+  int refinement_bits[kDCTBlockSize];
+  size_t refinement_bits_count = 0;
   for (int k = Ss; k <= Se; k++) {
     if (abs_values[k] == 0) {
       r++;
       continue;
     }
     while (r > 15 && k <= eob) {
-      Flush<kOutputMode>(coding_state, bw);
-      WriteSymbol<kOutputMode>(0xf0, ac_huff, bw);
+      Flush(coding_state, bw);
+      WriteSymbol(0xf0, ac_huff, bw);
       r -= 16;
-      for (int bit : refinement_bits) {
-        WriteBits(bw, 1, bit);
+      for (size_t i = 0; i < refinement_bits_count; ++i) {
+        WriteBits(bw, 1, refinement_bits[i]);
       }
-      refinement_bits.clear();
+      refinement_bits_count = 0;
     }
     if (abs_values[k] > 1) {
-      refinement_bits.push_back(abs_values[k] & 1u);
+      refinement_bits[refinement_bits_count++] = abs_values[k] & 1u;
       continue;
     }
-    Flush<kOutputMode>(coding_state, bw);
+    Flush(coding_state, bw);
     int symbol = (r << 4u) + 1;
     int new_non_zero_bit = (coeffs[kJPEGNaturalOrder[k]] < 0) ? 0 : 1;
-    WriteSymbol<kOutputMode>(symbol, ac_huff, bw);
+    WriteSymbol(symbol, ac_huff, bw);
     WriteBits(bw, 1, new_non_zero_bit);
-    for (int bit : refinement_bits) {
-      WriteBits(bw, 1, bit);
+    for (size_t i = 0; i < refinement_bits_count; ++i) {
+      WriteBits(bw, 1, refinement_bits[i]);
     }
-    refinement_bits.clear();
+    refinement_bits_count = 0;
     r = 0;
   }
-  if (r > 0 || !refinement_bits.empty()) {
-    BufferEndOfBand<kOutputMode>(coding_state, ac_huff, &refinement_bits, bw);
+  if (r > 0 || refinement_bits_count) {
+    BufferEndOfBand(coding_state, ac_huff, refinement_bits,
+                    refinement_bits_count, bw);
     if (!eob_run_allowed) {
-      Flush<kOutputMode>(coding_state, bw);
+      Flush(coding_state, bw);
     }
   }
   return true;
 }
 
-size_t NumHistograms(const JPEGData& jpg) {
-  size_t num = 0;
-  for (const auto& si : jpg.scan_info) {
-    num += si.num_components;
-  }
-  return num;
-}
-
-size_t HistogramIndex(const JPEGData& jpg, size_t scan_index,
-                      size_t component_index) {
-  size_t idx = 0;
-  for (size_t i = 0; i < scan_index; ++i) {
-    idx += jpg.scan_info[i].num_components;
-  }
-  return idx + component_index;
-}
-
-template <int kMode, int kOutputMode>
+template <int kMode>
 SerializationStatus JXL_NOINLINE DoEncodeScan(const JPEGData& jpg,
                                               SerializationState* state) {
   const JPEGScanInfo& scan_info = jpg.scan_info[state->scan_index];
@@ -748,7 +773,8 @@ SerializationStatus JXL_NOINLINE DoEncodeScan(const JPEGData& jpg,
 
   // DC-only is defined by [0..0] spectral range.
   const bool want_ac = ((Ss != 0) || (Se != 0));
-  // TODO: support streaming decoding again.
+  const bool want_dc = (Ss == 0);
+  // TODO(user): support streaming decoding again.
   const bool complete_ac = true;
   const bool has_ac = true;
   if (want_ac && !has_ac) return SerializationStatus::NEEDS_MORE_INPUT;
@@ -772,7 +798,7 @@ SerializationStatus JXL_NOINLINE DoEncodeScan(const JPEGData& jpg,
     for (int mcu_x = 0; mcu_x < MCUs_per_row; ++mcu_x) {
       // Possibly emit a restart marker.
       if (restart_interval > 0 && ss.restarts_to_go == 0) {
-        Flush<kOutputMode>(coding_state, bw);
+        Flush(coding_state, bw);
         if (!JumpToByteBoundary(bw, &state->pad_bits, state->pad_bits_end)) {
           return SerializationStatus::ERROR;
         }
@@ -782,18 +808,21 @@ SerializationStatus JXL_NOINLINE DoEncodeScan(const JPEGData& jpg,
         ss.restarts_to_go = restart_interval;
         memset(ss.last_dc_coeff, 0, sizeof(ss.last_dc_coeff));
       }
+
       // Encode one MCU
       for (size_t i = 0; i < scan_info.num_components; ++i) {
         const JPEGComponentScanInfo& si = scan_info.components[i];
         const JPEGComponent& c = jpg.components[si.comp_idx];
-        size_t dc_tbl_idx = (kOutputMode == OutputModes::kModeHistogram
-                                 ? HistogramIndex(jpg, state->scan_index, i)
-                                 : si.dc_tbl_idx);
-        size_t ac_tbl_idx = (kOutputMode == OutputModes::kModeHistogram
-                                 ? HistogramIndex(jpg, state->scan_index, i)
-                                 : si.ac_tbl_idx);
+        size_t dc_tbl_idx = si.dc_tbl_idx;
+        size_t ac_tbl_idx = si.ac_tbl_idx;
         HuffmanCodeTable* dc_huff = &state->dc_huff_table[dc_tbl_idx];
         HuffmanCodeTable* ac_huff = &state->ac_huff_table[ac_tbl_idx];
+        if (want_dc && !dc_huff->initialized) {
+          return SerializationStatus::ERROR;
+        }
+        if (want_ac && !ac_huff->initialized) {
+          return SerializationStatus::ERROR;
+        }
         int n_blocks_y = is_interleaved ? c.v_samp_factor : 1;
         int n_blocks_x = is_interleaved ? c.h_samp_factor : 1;
         for (int iy = 0; iy < n_blocks_y; ++iy) {
@@ -802,7 +831,7 @@ SerializationStatus JXL_NOINLINE DoEncodeScan(const JPEGData& jpg,
             int block_x = mcu_x * n_blocks_x + ix;
             int block_idx = block_y * c.width_in_blocks + block_x;
             if (ss.block_scan_index == ss.next_reset_point) {
-              Flush<kOutputMode>(coding_state, bw);
+              Flush(coding_state, bw);
               ss.next_reset_point = get_next_reset_point();
             }
             int num_zero_runs = 0;
@@ -814,17 +843,19 @@ SerializationStatus JXL_NOINLINE DoEncodeScan(const JPEGData& jpg,
             }
             const coeff_t* coeffs = &c.coeffs[block_idx << 6];
             bool ok;
+            // compressed size per block cannot be more than 512 bytes
+            Reserve(bw, 512);
             if (kMode == 0) {
-              ok = EncodeDCTBlockSequential<kOutputMode>(
-                  coeffs, dc_huff, ac_huff, num_zero_runs,
-                  ss.last_dc_coeff + si.comp_idx, bw);
+              ok = EncodeDCTBlockSequential(coeffs, dc_huff, ac_huff,
+                                            num_zero_runs,
+                                            ss.last_dc_coeff + si.comp_idx, bw);
             } else if (kMode == 1) {
-              ok = EncodeDCTBlockProgressive<kOutputMode>(
+              ok = EncodeDCTBlockProgressive(
                   coeffs, dc_huff, ac_huff, Ss, Se, Al, num_zero_runs,
                   coding_state, ss.last_dc_coeff + si.comp_idx, bw);
             } else {
-              ok = EncodeRefinementBits<kOutputMode>(coeffs, ac_huff, Ss, Se,
-                                                     Al, coding_state, bw);
+              ok = EncodeRefinementBits(coeffs, ac_huff, Ss, Se, Al,
+                                        coding_state, bw);
             }
             if (!ok) return SerializationStatus::ERROR;
             ++ss.block_scan_index;
@@ -838,7 +869,7 @@ SerializationStatus JXL_NOINLINE DoEncodeScan(const JPEGData& jpg,
     if (!bw->healthy) return SerializationStatus::ERROR;
     return SerializationStatus::NEEDS_MORE_INPUT;
   }
-  Flush<kOutputMode>(coding_state, bw);
+  Flush(coding_state, bw);
   if (!JumpToByteBoundary(bw, &state->pad_bits, state->pad_bits_end)) {
     return SerializationStatus::ERROR;
   }
@@ -850,9 +881,8 @@ SerializationStatus JXL_NOINLINE DoEncodeScan(const JPEGData& jpg,
   return SerializationStatus::DONE;
 }
 
-template <int kOutputMode>
-static SerializationStatus JXL_INLINE EncodeScan(const JPEGData& jpg,
-                                                 SerializationState* state) {
+SerializationStatus JXL_INLINE EncodeScan(const JPEGData& jpg,
+                                          SerializationState* state) {
   const JPEGScanInfo& scan_info = jpg.scan_info[state->scan_index];
   const bool is_progressive = state->is_progressive;
   const int Al = is_progressive ? scan_info.Al : 0;
@@ -862,15 +892,14 @@ static SerializationStatus JXL_INLINE EncodeScan(const JPEGData& jpg,
   const bool need_sequential =
       !is_progressive || (Ah == 0 && Al == 0 && Ss == 0 && Se == 63);
   if (need_sequential) {
-    return DoEncodeScan<0, kOutputMode>(jpg, state);
+    return DoEncodeScan<0>(jpg, state);
   } else if (Ah == 0) {
-    return DoEncodeScan<1, kOutputMode>(jpg, state);
+    return DoEncodeScan<1>(jpg, state);
   } else {
-    return DoEncodeScan<2, kOutputMode>(jpg, state);
+    return DoEncodeScan<2>(jpg, state);
   }
 }
 
-template <int kOutputMode>
 SerializationStatus SerializeSection(uint8_t marker, SerializationState* state,
                                      const JPEGData& jpg) {
   const auto to_status = [](bool result) {
@@ -886,8 +915,7 @@ SerializationStatus SerializeSection(uint8_t marker, SerializationState* state,
       return to_status(EncodeSOF(jpg, marker, state));
 
     case 0xC4:
-      return to_status((kOutputMode == OutputModes::kModeHistogram) ||
-                       EncodeDHT(jpg, state));
+      return to_status(EncodeDHT(jpg, state));
 
     case 0xD0:
     case 0xD1:
@@ -903,7 +931,7 @@ SerializationStatus SerializeSection(uint8_t marker, SerializationState* state,
       return to_status(EncodeEOI(jpg, state));
 
     case 0xDA:
-      return EncodeScan<kOutputMode>(jpg, state);
+      return EncodeScan(jpg, state);
 
     case 0xDB:
       return to_status(EncodeDQT(jpg, state));
@@ -941,7 +969,6 @@ SerializationStatus SerializeSection(uint8_t marker, SerializationState* state,
 }
 
 // TODO(veluca): add streaming support again.
-template <int kOutputMode>
 Status WriteJpegInternal(const JPEGData& jpg, const JPEGOutput& out,
                          SerializationState* ss) {
   const auto maybe_push_output = [&]() -> Status {
@@ -972,18 +999,8 @@ Status WriteJpegInternal(const JPEGData& jpg, const JPEGOutput& out,
           ss->stage = SerializationState::STAGE_ERROR;
           break;
         }
-        if (kOutputMode == OutputModes::kModeHistogram) {
-          size_t num_histo = NumHistograms(jpg);
-          ss->dc_huff_table.resize(num_histo);
-          ss->ac_huff_table.resize(num_histo);
-          for (size_t i = 0; i < num_histo; ++i) {
-            ss->dc_huff_table[i].InitDepths();
-            ss->ac_huff_table[i].InitDepths();
-          }
-        } else {
-          ss->dc_huff_table.resize(kMaxHuffmanTables);
-          ss->ac_huff_table.resize(kMaxHuffmanTables);
-        }
+        ss->dc_huff_table.resize(kMaxHuffmanTables);
+        ss->ac_huff_table.resize(kMaxHuffmanTables);
         if (jpg.has_zero_padding_bit) {
           ss->pad_bits = jpg.padding_bits.data();
           ss->pad_bits_end = ss->pad_bits + jpg.padding_bits.size();
@@ -1001,8 +1018,7 @@ Status WriteJpegInternal(const JPEGData& jpg, const JPEGOutput& out,
           break;
         }
         uint8_t marker = jpg.marker_order[ss->section_index];
-        SerializationStatus status =
-            SerializeSection<kOutputMode>(marker, ss, jpg);
+        SerializationStatus status = SerializeSection(marker, ss, jpg);
         if (status == SerializationStatus::ERROR) {
           JXL_WARNING("Failed to encode marker 0x%.2x", marker);
           ss->stage = SerializationState::STAGE_ERROR;
@@ -1036,21 +1052,8 @@ Status WriteJpegInternal(const JPEGData& jpg, const JPEGOutput& out,
 }  // namespace
 
 Status WriteJpeg(const JPEGData& jpg, const JPEGOutput& out) {
-  SerializationState ss;
-  return WriteJpegInternal<OutputModes::kModeWrite>(jpg, out, &ss);
-}
-
-Status ProcessJpeg(const JPEGData& jpg, SerializationState* ss) {
-  auto nullout = [](const uint8_t* buf, size_t len) { return len; };
-  return WriteJpegInternal<OutputModes::kModeHistogram>(jpg, nullout, ss);
-}
-
-Status EncodeImageJPGCoefficients(const CodecInOut* io, PaddedBytes* bytes) {
-  auto write = [&bytes](const uint8_t* buf, size_t len) {
-    bytes->append(buf, buf + len);
-    return len;
-  };
-  return WriteJpeg(*io->Main().jpeg_data, write);
+  auto ss = jxl::make_unique<SerializationState>();
+  return WriteJpegInternal(jpg, out, ss.get());
 }
 
 }  // namespace jpeg
