@@ -13,14 +13,12 @@
 
 #include <jxl/codestream_header.h>
 #include <jxl/encode.h>
-#include <jxl/encode_cxx.h>
 #include <jxl/thread_parallel_runner.h>
 #include <jxl/thread_parallel_runner_cxx.h>
 #include <jxl/types.h>
 
 #include <algorithm>
 #include <cerrno>
-#include <cmath>
 #include <cstdint>
 #include <cstdio>
 #include <cstdlib>
@@ -28,24 +26,23 @@
 #include <functional>
 #include <iostream>
 #include <memory>
-#include <sstream>
 #include <string>
-#include <thread>
-#include <type_traits>
+#include <utility>
 #include <vector>
 
-#include "lib/extras/dec/apng.h"
 #include "lib/extras/dec/color_hints.h"
 #include "lib/extras/dec/decode.h"
 #include "lib/extras/dec/pnm.h"
 #include "lib/extras/enc/jxl.h"
+#include "lib/extras/packed_image.h"
 #include "lib/extras/time.h"
+#include "lib/jxl/base/c_callback_support.h"
 #include "lib/jxl/base/common.h"
+#include "lib/jxl/base/exif.h"
 #include "lib/jxl/base/override.h"
 #include "lib/jxl/base/printf_macros.h"
 #include "lib/jxl/base/span.h"
 #include "lib/jxl/base/status.h"
-#include "lib/jxl/exif.h"
 #include "tools/args.h"
 #include "tools/cmdline.h"
 #include "tools/codec_config.h"
@@ -154,7 +151,7 @@ struct CompressArgs {
         "Target visual distance for the alpha channel, lower = higher "
         "quality.\n"
         "    0.0 = mathematically lossless. 1.0 = visually lossless.\n"
-        "    Default is to use the same value as for the color image.\n"
+        "    Default is 0.\n"
         "    Recommended range: 0.5 .. 3.0. Allowed range: 0.0 ... 25.0.",
         &alpha_distance, &ParseFloat, 1);
 
@@ -237,17 +234,20 @@ struct CompressArgs {
         "    The keys 'exif', 'xmp', and 'jumbf' refer to a binary file "
         "containing metadata;\n"
         "    existing metadata of the same type will be overwritten.\n"
-        "    Specific metadata can be stripped using e.g. -x strip=exif",
+        "    Specific metadata can be stripped using e.g. -x strip=exif."
+        "    Stripping metadata when losslessly recompression JPEGs only works "
+        "    without reconstruction, hence `--allow_jpeg_reconstruction=0` "
+        "    must be passed in this case.",
         &color_hints_proxy, &ParseAndAppendKeyValue<ColorHintsProxy>, 1);
 
     cmdline->AddHelpText("\nExpert options:", 2);
 
     cmdline->AddOptionValue(
-        '\0', "jpeg_store_metadata", "0|1",
+        '\0', "allow_jpeg_reconstruction", "0|1",
         ("If --lossless_jpeg=1, store JPEG reconstruction "
          "metadata in the JPEG XL container.\n"
          "    This allows reconstruction of the JPEG codestream. Default: 1."),
-        &jpeg_store_metadata, &ParseUnsigned, 2);
+        &allow_jpeg_reconstruction, &ParseUnsigned, 2);
 
     cmdline->AddOptionValue('\0', "codestream_level", "K",
                             "The codestream level. Either `-1`, `5` or `10`.",
@@ -389,7 +389,7 @@ struct CompressArgs {
 
     cmdline->AddOptionFlag('\0', "allow_expert_options",
                            "Allow specifying advanced options; this allows "
-                           "setting effort to 10, for\n"
+                           "setting effort to 11, for\n"
                            "    somewhat better lossless compression at the "
                            "cost of a massive speed hit.",
                            &allow_expert_options, &SetBooleanTrue, 3);
@@ -423,7 +423,7 @@ struct CompressArgs {
         "    5=gradient, 6=weighted, 7=topright, 8=topleft, 9=leftleft, "
         "10=avg1, 11=avg2, 12=avg3,\n"
         "    13=toptop predictive average, 14=mix 5 and 6, 15=mix everything.\n"
-        "    Default is 14 at effort < 9 and 15 at effort 9.",
+        "    Default is 14 at effort < 9 and 15 at effort 9-10.",
         &modular_predictor, &ParseInt64, 4);
 
     cmdline->AddOptionValue(
@@ -490,7 +490,7 @@ struct CompressArgs {
   // Reset to false if input image is not a JPEG.
   size_t lossless_jpeg = 1;
 
-  size_t jpeg_store_metadata = 1;
+  size_t allow_jpeg_reconstruction = 1;
 
   float quality = -1001.f;  // Default to lossless if input is already lossy,
                             // or to VarDCT otherwise.
@@ -585,17 +585,20 @@ void PrintMode(jxl::extras::PackedPixelFile& ppf, const double decode_mps,
       (args.container == jxl::Override::kOn ? "Container | " : ""), mode,
       distance.c_str(), args.effort);
   if (args.container == jxl::Override::kOn) {
-    if (args.lossless_jpeg && args.jpeg_store_metadata)
+    if (args.lossless_jpeg && args.allow_jpeg_reconstruction)
       cmdline.VerbosePrintf(0, " | JPEG reconstruction data");
-    if (!ppf.metadata.exif.empty())
+    if (!ppf.metadata.exif.empty()) {
       cmdline.VerbosePrintf(0, " | %" PRIuS "-byte Exif",
                             ppf.metadata.exif.size());
-    if (!ppf.metadata.xmp.empty())
+    }
+    if (!ppf.metadata.xmp.empty()) {
       cmdline.VerbosePrintf(0, " | %" PRIuS "-byte XMP",
                             ppf.metadata.xmp.size());
-    if (!ppf.metadata.jumbf.empty())
+    }
+    if (!ppf.metadata.jumbf.empty()) {
       cmdline.VerbosePrintf(0, " | %" PRIuS "-byte JUMBF",
                             ppf.metadata.jumbf.size());
+    }
   }
   cmdline.VerbosePrintf(0, "]\n");
 }
@@ -670,8 +673,7 @@ void SetDistanceFromFlags(CommandLineParser* cmdline, CompressArgs* args,
     args->lossless_jpeg = 0;
   }
   params->distance = args->distance;
-  params->alpha_distance =
-      alpha_distance_set ? args->alpha_distance : params->distance;
+  params->alpha_distance = alpha_distance_set ? args->alpha_distance : 0;
 }
 
 void ProcessFlags(const jxl::extras::Codec codec,
@@ -714,9 +716,9 @@ void ProcessFlags(const jxl::extras::Codec codec,
       "effort", static_cast<int64_t>(args->effort),
       JXL_ENC_FRAME_SETTING_EFFORT, params, [args](int64_t x) -> std::string {
         if (args->allow_expert_options) {
-          return (1 <= x && x <= 10) ? "" : "Valid range is {1, 2, ..., 10}.";
+          return (1 <= x && x <= 11) ? "" : "Valid range is {1, 2, ..., 11}.";
         } else {
-          return (1 <= x && x <= 9) ? "" : "Valid range is {1, 2, ..., 9}.";
+          return (1 <= x && x <= 10) ? "" : "Valid range is {1, 2, ..., 10}.";
         }
       });
   ProcessFlag("brotli_effort", static_cast<int64_t>(args->brotli_effort),
@@ -913,7 +915,7 @@ void ProcessFlags(const jxl::extras::Codec codec,
   }
   // Copy over the rest of the non-option params.
   params->use_container = args->container == jxl::Override::kOn;
-  params->jpeg_store_metadata = args->jpeg_store_metadata;
+  params->jpeg_store_metadata = args->allow_jpeg_reconstruction;
   params->intensity_target = args->intensity_target;
   params->override_bitdepth = args->override_bitdepth;
   params->codestream_level = args->codestream_level;
@@ -930,7 +932,7 @@ void ProcessFlags(const jxl::extras::Codec codec,
   (void)args->color_hints_proxy.target.Foreach(
       [&params](const std::string& key,
                 const std::string& value) -> jxl::Status {
-        if (value == "") {
+        if (value.empty()) {
           if (key == "exif") params->jpeg_strip_exif = true;
           if (key == "xmp") params->jpeg_strip_xmp = true;
           if (key == "jumbf") params->jpeg_strip_jumbf = true;
@@ -953,35 +955,37 @@ struct JxlOutputProcessor {
   }
 
   JxlEncoderOutputProcessor GetOutputProcessor() {
-    return JxlEncoderOutputProcessor{this, GetBuffer, ReleaseBuffer, Seek,
-                                     SetFinalizedPosition};
+    return JxlEncoderOutputProcessor{
+        this, METHOD_TO_C_CALLBACK(&JxlOutputProcessor::GetBuffer),
+        METHOD_TO_C_CALLBACK(&JxlOutputProcessor::ReleaseBuffer),
+        METHOD_TO_C_CALLBACK(&JxlOutputProcessor::Seek),
+        METHOD_TO_C_CALLBACK(&JxlOutputProcessor::SetFinalizedPosition)};
   }
 
-  static void* GetBuffer(void* opaque, size_t* size) {
-    JxlOutputProcessor* self = reinterpret_cast<JxlOutputProcessor*>(opaque);
-    self->output.resize(*size);
-    return self->output.data();
+  void* GetBuffer(size_t* size) {
+    *size = std::min<size_t>(*size, 1u << 16);
+    if (output.size() < *size) {
+      output.resize(*size);
+    }
+    return output.data();
   }
 
-  static void ReleaseBuffer(void* opaque, size_t written_bytes) {
-    JxlOutputProcessor* self = reinterpret_cast<JxlOutputProcessor*>(opaque);
-    if (*self->outfile && fwrite(self->output.data(), 1, written_bytes,
-                                 *self->outfile) != written_bytes) {
+  void ReleaseBuffer(size_t written_bytes) {
+    if (*outfile &&
+        fwrite(output.data(), 1, written_bytes, *outfile) != written_bytes) {
       JXL_WARNING("Failed to write %" PRIuS " bytes to output", written_bytes);
     }
-    self->output.clear();
+    output.clear();
   }
 
-  static void Seek(void* opaque, uint64_t position) {
-    JxlOutputProcessor* self = reinterpret_cast<JxlOutputProcessor*>(opaque);
-    if (*self->outfile && fseek(*self->outfile, position, SEEK_SET) != 0) {
+  void Seek(uint64_t position) {
+    if (*outfile && fseek(*outfile, position, SEEK_SET) != 0) {
       JXL_WARNING("Failed to seek output.");
     }
   }
 
-  static void SetFinalizedPosition(void* opaque, uint64_t finalized_position) {
-    JxlOutputProcessor* self = reinterpret_cast<JxlOutputProcessor*>(opaque);
-    self->finalized_position = finalized_position;
+  void SetFinalizedPosition(uint64_t finalized_position) {
+    this->finalized_position = finalized_position;
   }
 
   std::vector<uint8_t> output;
@@ -1031,34 +1035,37 @@ int main(int argc, char** argv) {
             "Encoding will be performed, but the result will be discarded.\n");
   }
 
-  // Loading the input.
-  // Depending on flags-settings, we want to either load a JPEG and
-  // faithfully convert it to JPEG XL, or load (JPEG or non-JPEG)
-  // pixel data.
-  jpegxl::tools::FileWrapper f(args.file_in, "rb");
-  if (!f) {
-    std::cerr << "Reading image data failed." << std::endl;
-    exit(EXIT_FAILURE);
-  }
   jxl::extras::JXLCompressParams params;
   jxl::extras::PackedPixelFile ppf;
   jxl::extras::Codec codec = jxl::extras::Codec::kUnknown;
   std::vector<uint8_t> image_data;
   std::vector<uint8_t>* jpeg_bytes = nullptr;
-  jxl::extras::ChunkedPNMDecoder pnm_dec;
   size_t input_bytes = 0;
   double decode_mps = 0;
   size_t pixels = 0;
+  jxl::extras::ChunkedPNMDecoder pnm_dec;
   if (args.streaming_input) {
-    pnm_dec.f = f;
-    if (!DecodeImagePNM(&pnm_dec, args.color_hints_proxy.target, &ppf)) {
+    auto dec = jxl::extras::ChunkedPNMDecoder::Init(args.file_in);
+    if (!dec.ok()) {
       std::cerr << "PNM decoding failed." << std::endl;
       exit(EXIT_FAILURE);
     }
+    pnm_dec = std::move(dec).value();
+    JXL_RETURN_IF_ERROR(
+        pnm_dec.InitializePPF(args.color_hints_proxy.target, &ppf));
     codec = jxl::extras::Codec::kPNM;
     args.lossless_jpeg = 0;
     pixels = ppf.info.xsize * ppf.info.ysize;
   } else {
+    // Loading the input.
+    // Depending on flags-settings, we want to either load a JPEG and
+    // faithfully convert it to JPEG XL, or load (JPEG or non-JPEG)
+    // pixel data.
+    jpegxl::tools::FileWrapper f(args.file_in, "rb");
+    if (!f) {
+      std::cerr << "Reading image data failed." << std::endl;
+      exit(EXIT_FAILURE);
+    }
     if (!jpegxl::tools::ReadFile(f, &image_data)) {
       std::cerr << "Reading image data failed." << std::endl;
       exit(EXIT_FAILURE);
@@ -1092,6 +1099,24 @@ int main(int argc, char** argv) {
                   << std::endl;
       }
       jpeg_bytes = &image_data;
+      if (args.allow_jpeg_reconstruction) {
+        (void)args.color_hints_proxy.target.Foreach([](const std::string& key,
+                                                       const std::string& value)
+                                                        -> jxl::Status {
+          if (value.empty()) {
+            if (key != "jumbf") {
+              std::cerr
+                  << "Cannot strip " << key
+                  << " metadata, try setting --allow_jpeg_reconstruction=0. "
+                     "Note that with that setting byte exact reconstruction "
+                     "of the JPEG file won't be possible."
+                  << std::endl;
+              exit(EXIT_FAILURE);
+            }
+          }
+          return true;
+        });
+      }
     }
   }
 
@@ -1107,7 +1132,7 @@ int main(int argc, char** argv) {
 
   if (!ppf.metadata.exif.empty() || !ppf.metadata.xmp.empty() ||
       !ppf.metadata.jumbf.empty() || !ppf.metadata.iptc.empty() ||
-      (args.lossless_jpeg && args.jpeg_store_metadata)) {
+      (args.lossless_jpeg && args.allow_jpeg_reconstruction)) {
     if (args.container == jxl::Override::kDefault) {
       args.container = jxl::Override::kOn;
     } else if (args.container == jxl::Override::kOff) {
@@ -1117,7 +1142,7 @@ int main(int argc, char** argv) {
       ppf.metadata.xmp.clear();
       ppf.metadata.jumbf.clear();
       ppf.metadata.iptc.clear();
-      args.jpeg_store_metadata = 0;
+      args.allow_jpeg_reconstruction = 0;
     }
   }
 
@@ -1131,7 +1156,7 @@ int main(int argc, char** argv) {
   params.runner = JxlThreadParallelRunner;
   params.runner_opaque = runner.get();
 
-  if (args.streaming_input && args.streaming_output) {
+  if (args.streaming_input) {
     params.options.emplace_back(jxl::extras::JXLOption(
         JXL_ENC_FRAME_SETTING_BUFFERING, static_cast<int64_t>(3), 0));
   }
@@ -1147,6 +1172,10 @@ int main(int argc, char** argv) {
   }
   std::vector<uint8_t> compressed;
   for (size_t num_rep = 0; num_rep < args.num_reps; ++num_rep) {
+    if (args.streaming_output) {
+      output_processor.Seek(0);
+      output_processor.SetFinalizedPosition(0);
+    }
     const double t0 = jxl::Now();
     if (!EncodeImageJXL(params, ppf, jpeg_bytes,
                         args.streaming_output ? nullptr : &compressed)) {
